unit aibrain;
	{ This unit handles behavior for the various enemy units in }
	{ Gearhead Arena. }

{
	GearHead: Arena, a roguelike mecha CRPG
	Copyright (C) 2005 Joseph Hewitt

	This library is free software; you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation; either version 2.1 of the License, or (at
	your option) any later version.

	The full text of the LGPL can be found in license.txt.

	This library is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
	General Public License for more details. 

	You should have received a copy of the GNU Lesser General Public License
	along with this library; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
}

interface

uses
{$IFDEF PATCH_GH}
	gears_base,
{$ENDIF PATCH_GH}
	gears,locale;

Procedure ClearHotMaps;

Procedure GetAIInput( Mek: GearPtr; GB: GameBoardPtr );
Procedure ConfusedInput( Mek: GearPtr; GB: GameBoardPtr );

Procedure BrownianMotion( GB: GameBoardPtr );

Procedure LancemateUsefulAction( Mek: GearPtr; GB: GameBoardPtr );

Function MOVE_MODEL_TOWARDS_SPOT( Mek: GearPtr; GB: GameBoardPtr; GX,GY: Integer ): Boolean;

{$IFDEF PATCH_HEART}
Procedure NPC_CombatTaunt( GB: GameBoardPtr; NPC: GearPtr; Msg_Label: String );
procedure AIAttacker( GB: GameBoardPtr; Mek,Weapon,Target: GearPtr );
Function SelectBestWeapon( GB: GameBoardPtr; Mek,Target: GearPtr ): GearPtr;
Function DoorPresent( GB: GameBoardPtr; X,Y: Integer ): Boolean;
Function AIRepair( GB: GameBoardPtr; NPC,Target: GearPtr; RepairLimit: LongInt; Skill: Integer ): LongInt;
Procedure NPC_Flirtation( GB: GameBoardPtr; NPC , Target: GearPtr );
Procedure NPC_Chatting( GB: GameBoardPtr; NPC , Target: GearPtr );
Procedure AI_Eject( Mek: GearPtr; GB: GameBoardPtr );
function SafeToFire( GB: GameBoardPtr; Mek, Part, Target: GearPtr ): Boolean;
function WGoodness( GB: GameBoardPtr; Part, Target: GearPtr ): Integer;
{$ENDIF PATCH_HEART}

implementation

uses
{$IFDEF DEBUG}
	errmsg,
{$ELSE DEBUG}
  {$IFDEF PATCH_HEART}
	errmsg,
  {$ENDIF PATCH_HEART}
{$ENDIF DEBUG}
{$IFDEF PATCH_I18N}
	i18nmsg,
{$ENDIF PATCH_I18N}
{$IFDEF PATCH_GH}
	ui4gh,
	ability,action,arenacfe,backpack,damage,effects,movement,gearutil,
	ghchars,ghmodule,ghweapon,ghparser,ghprop,interact,rpgdice,skilluse,
	texutil,
{$ELSE PATCH_GH}
	ability,action,arenacfe,damage,effects,movement,gearutil,
	ghchars,ghmodule,ghweapon,ghparser,ghprop,interact,rpgdice,skilluse,
	texutil,ui4gh,
{$ENDIF PATCH_GH}
{$IFDEF SDLMODE}
	sdlmap,sdlgfx
{$ELSE SDLMODE}
	conmap,context
{$ENDIF SDLMODE}
{$IFDEF PATCH_CHEAT}
	,pcaction,ghmecha
{$ENDIF PATCH_CHEAT}
	;

const
	Hot_Terr: Array [0..NumMoveMode,1..NumTerr] of SmallInt = (
	( 2, 3, 4,-1, 2,  2, 3, 3, 4, 5,  { CHARA WALK }
	  3,-1,-1, 2,-1,  2, 2,-1, 2, 2,
	 -1,-1,-1,-1, 2,  2,-1, 2,-1, 2,
	 -1,-1,-1,-1,-1, -1,-1,-1, 2, 4,
	  2,-1	),
	( 2, 3, 4, 5, 2,  2, 3, 3, 4, 5,  { MECHA WALK }
	  3,-1,-1, 2,-1,  2, 2,-1, 2, 2,
	  5, 5,-1,-1, 2,  2,-1, 2,-1, 2,
	 -1,-1,-1,-1,-1, -1,-1,-1, 2, 4,
	  2,-1	),
	( 2,-1,-1,-1, 3,  1, 4, 3, 4, 5,  { ROLL }
	  4,-1,-1, 2,-1,  2, 2,-1, 2, 2,
	  5, 5,-1,-1, 2,  2,-1, 2,-1, 2,
	 -1,-1,-1,-1,-1, -1,-1,-1, 2,-1,
	  2,-1	),
	( 2, 3, 4, 2, 2,  2, 3, 3, 4, 5,  { SKIM }
	  2,-1,-1, 2,-1,  2, 2,-1, 2, 2,
	  2, 2,-1,-1, 2,  2,-1, 2,-1, 2,
	 -1,-1,-1,-1,-1, -1,-1,-1, 2, 4,
	  2,-1	),
	( 2, 2, 2, 2, 2,  2, 2, 2, 2, 2,  { FLY }
	  2, 2,-1, 2,-1,  2, 2,-1, 2, 2,
	  2, 2,-1,-1, 2,  2,-1, 2,-1, 2,
	 -1,-1,-1,-1,-1, -1,-1,-1, 2, 2,
	  2,-1	)
	);

	NumFFMap = 10;

	ORD_SeekEnemy = 0;
	ORD_SeekSingleModel = 1;
	ORD_SeekSpot = 2;
	ORD_SeekEdge = 3;
	ORD_SeekTeam = 4;
	ORD_SeekNearestEdge = 5;

	Hot_Map_Validity = 10;

	CORD_Flirt = 1;
	CORD_Chat = 2;

{$IFDEF PATCH_GH}
	Hot_Map_UpperLimit = 9999;
{$ENDIF PATCH_GH}

var
	{ The HOTMAP is used for pathfinding and general tactical }
	{ decision-making. }
	{ HOTMAP shows what squares the NPC will want to move into. }
	{ COLDMAP shows what squares the NPC won't want to move into. }
	MapUpdate: Array [1..NumFFMap] of LongInt;
	MapTeam: Array [1..NumFFMap] of Integer;
	MapOrder: Array [1..NumFFMap] of Integer;
	MapMoveMode: Array [1..NumFFMap] of Integer;
	HotMap: Array [1..NumFFMap , 1..XMax , 1..YMax ] of Integer;
	ColdMap: Array [1..NumFFMap , 1..XMax , 1..YMax ] of Integer;

	NPC_Chatter_Standard: SAttPtr;

Procedure NPC_CombatTaunt( GB: GameBoardPtr; NPC: GearPtr; Msg_Label: String );
	{ NPC is going to say something... maybe. Search the NPC gear for things to say, }
	{ then search the standard chatter list for things to say, then pick one of them }
	{ and say it. }
	{ Note that this will not nessecarily be an actual taunt. It's more likely to be }
	{ something completely different... but I didn't feel like calling this procedure }
	{ "NPC_Mutters"... }
var
	MList: SAttPtr;
	Procedure HarvestMessages( LList: SAttPtr; head: String );
		{ Look through LList and collect all strings that match HEAD. }
	var
		M: SAttPtr;
	begin
		M := LList;
		while M <> Nil do begin
			if HeadMatchesString( head , M^.Info ) then StoreSAtt( MList , RetrieveAString( M^.Info ) );
			M := M^.Next;
		end;
	end;
var
	T,V: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = NPC) or (NPC^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Make sure combat taunts are enabled. }
	if No_Combat_Taunts then begin
		SetNAtt( FindRoot( NPC )^.NA , NAG_EpisodeData , NAS_ChatterRecharge , GB^.ComTime + 100000 );
		Exit;
	end;

	{ Make sure we have a proper NPC, and not a mecha. }
	NPC := LocatePilot( NPC );
{$IFDEF PATCH_GH}
	if (NIL = NPC) or (NPC^.G <= GG_DisposeGear) then Exit;
{$ELSE PATCH_GH}
	if NPC = Nil then Exit;
{$ENDIF PATCH_GH}

	{ Initialize our message list to NIL. }
	MList := Nil;

	{ First, search through NPC itself looking for appropriate messages. }
	HarvestMessages( NPC^.SA , msg_label );

	{ Next, pick some contenders from the standard chatter list. }
	{ Only characters with CIDs get this. }
	if ( UpCase( SAttValue( NPC^.SA , 'JOB' ) ) <> 'ANIMAL' ) and ( ( Msg_Label = 'CHAT_EJECT' ) or ( ( NAttValue( NPC^.NA , NAG_Personal , NAS_CID ) <> 0 ) and ( ( MList = Nil ) or ( Random( 2 ) = 1 ) ) ) ) then begin
		HarvestMessages( NPC_Chatter_Standard , msg_label + '_ALL' );

		for t := 1 to Num_Personality_Traits do begin
			V := NAttValue( NPC^.NA , NAG_CharDescription , -T );
			if V > 0 then begin
				HarvestMessages( NPC_Chatter_Standard , msg_label + '_T' + BStr( T ) + '+' );
			end else if V < 0 then begin
				HarvestMessages( NPC_Chatter_Standard , msg_label + '_T' + BStr( T ) + '-' );
			end;
		end;
	end;

	{ If at least one phrase was found, and the NPC is visible, it can say something. }
	if ( MList <> Nil ) and ( ( Msg_Label = 'CHAT_EJECT' ) or MekVisible( GB , FindRoot( NPC ) ) ) then begin
{$IFDEF PATCH_I18N}
		DialogMsg( FormatChatStringByGender( '[' + GearName( NPC ) + ']: ' + SelectRandomSAtt( MList )^.Info, NPC ) );
{$ELSE PATCH_I18N}
		DialogMsg( '[' + GearName( NPC ) + ']: ' + SelectRandomSAtt( MList )^.Info );
{$ENDIF PATCH_I18N}
	end;

	{ Add the chatter recharge time. }
	SetNAtt( FindRoot( NPC )^.NA , NAG_EpisodeData , NAS_ChatterRecharge , GB^.ComTime + ( 2500 div CStat( NPC , STAT_Charm ) ) );

	{ Get rid of our message list. }
	DisposeSAtt( MList );
end;


Function HotTileBlocksLOS( GB: GameBoardPtr; X , Y , MM: Integer ): Boolean;
	{ Return TRUE if the given tile should block AI movement. }
	{ BUGS: X,Y must lie on the map; MM must be in range 0..NumMoveMode. }
begin
	HotTileBlocksLOS := TileBlocksLOS( GB , X , Y , 5 ) or ( Hot_Terr[ MM , GB^.Map[ X , Y ].Terr ] < 0 );
end;

Procedure HotMapFloodFill( GB: GameBoardPtr; N,MM: Integer );
	{ Given a map that's had the hot spots filled in, use the }
	{ flood fill algorithm to determine paths etc. }
var
	X,Y: Integer;
	Flag: Boolean;
	DV,DH,DD,DP: Integer;	{ Distance Vertical, Horizontal, and Diagnol. }
begin
	flag := True;
	while flag do begin
		flag := False;
{$IFDEF PATCH_GH}
		for y := 1 to YMax do begin
			for x := 1 to XMax do begin
				if not HotTileBlocksLOS( GB , X , Y , MM ) then begin
					dH := Hot_Map_UpperLimit;
					if OnTheMap( X - 1 , Y ) then begin
						dH := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X - 1 , Y ];
					end;
					dV := Hot_Map_UpperLimit;
					if OnTheMap( X , Y - 1 ) then begin
						dV := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X , Y - 1 ];
					end;
					if DV < DH then DH := DV;
					DD := Hot_Map_UpperLimit;
					if OnTheMap( X - 1 , Y - 1 ) then begin
						DD := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X - 1 , Y - 1 ];
					end;
					if DD < DH then DH := DD;
					DP := Hot_Map_UpperLimit;
					if OnTheMap( X + 1 , Y - 1 ) then begin
						DP := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X + 1 , Y - 1 ];
					end;
					if DP < DH then DH := DP;

					if DH < HotMap[ N , X , Y ] then begin
						HotMap[ N , X , Y ] := DH;
						flag := True;
					end;
				end;
			end;
		end;


		for y := YMax downto 1 do begin
			for x := XMax downto 1 do begin
				if not HotTileBlocksLOS( GB , X , Y , MM ) then begin
					dH := Hot_Map_UpperLimit;
					if OnTheMap( X + 1 , Y ) then begin
						dH := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X + 1 , Y ];
					end;
					dV := Hot_Map_UpperLimit;
					if OnTheMap( X , Y + 1 ) then begin
						dV := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X , Y + 1 ];
					end;
					if DV < DH then DH := DV;
					DD := Hot_Map_UpperLimit;
					if OnTheMap( X + 1 , Y + 1 ) then begin
						DD := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X + 1 , Y + 1 ];
					end;
					if DD < DH then DH := DD;
					DP := Hot_Map_UpperLimit;
					if OnTheMap( X - 1 , Y + 1 ) then begin
						DP := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X - 1 , Y + 1 ];
					end;
					if DP < DH then DH := DP;

					if DH < HotMap[ N , X , Y ] then begin
						HotMap[ N , X , Y ] := DH;
						flag := True;
					end;
				end;
			end;
		end;
{$ELSE PATCH_GH}
		for y := 2 to (YMax-1) do begin
			for x := 2 to (XMax-1) do begin
				if not HotTileBlocksLOS( GB , X , Y , MM ) then begin
					dH := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X - 1 , Y ];
					dV := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X , Y - 1 ];
					if DV < DH then DH := DV;
					DD := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X - 1 , Y - 1];
					if DD < DH then DH := DD;
					DP := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X + 1 , Y - 1];
					if DP < DH then DH := DP;

					if DH < HotMap[ N , X , Y ] then begin
						HotMap[ N , X , Y ] := DH;
						flag := True;
					end;
				end;
			end;
		end;


		for y := ( YMax - 1 ) downto 2 do begin
			for x := ( XMax - 1 ) downto 2 do begin
				if not HotTileBlocksLOS( GB , X , Y , MM ) then begin
					dH := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X + 1 , Y ];
					dV := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + HotMap[ N , X , Y + 1 ];
					if DV < DH then DH := DV;
					DD := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X + 1 , Y + 1];
					if DD < DH then DH := DD;
					DP := Hot_Terr[ MM , GB^.Map[X,Y].terr ] + 1 + HotMap[ N , X - 1 , Y + 1];
					if DP < DH then DH := DP;

					if DH < HotMap[ N , X , Y ] then begin
						HotMap[ N , X , Y ] := DH;
						flag := True;
					end;
				end;
			end;
		end;
{$ENDIF PATCH_GH}

	end;
end;

Procedure Calc_FollowMap( GB: GameBoardPtr; N,UID,MM: Integer );
	{ Unlike the SEEK AND DESTROY map, this map should have only }
	{ one hot spot. }
var
	M: GearPtr;
	P: Point;
begin
	M := GB^.Meks;
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if IsMasterGear( M ) and OnTheMap( P.X , P.Y ) and GearOperational( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end else if ( M^.G = GG_MetaTerrain ) and ( M^.Stat[ STAT_Pass ] <= -100 ) and NotDestroyed( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end;
		if NAttValue( M^.NA , NAG_EpisodeData , NAS_UID ) = UID then begin
			P := GearCurrentLocation( M );
			if OnTheMap( P.X , P.Y ) then begin
				HotMap[ N , P.X , P.Y ] := 0;
				HotMapFloodFill( GB , N , MM );
			end;
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;
end;

Procedure Calc_SpotMap( GB: GameBoardPtr; N,PDat,MM: Integer );
	{ Calculate a map seeking the described point. }
var
	M: GearPtr;
	P: Point;
begin
	{ Set the position for all blocking gears on the coldmap. }
	M := GB^.Meks;
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if IsMasterGear( M ) and OnTheMap( P.X , P.Y ) and GearOperational( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end else if ( M^.G = GG_MetaTerrain ) and ( M^.Stat[ STAT_Pass ] <= -100 ) and NotDestroyed( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;

	{ Set the hot spot. }
	P.X := PDat mod ( XMax + 1 );
	P.Y := PDat div ( XMax + 1 );
	if OnTheMap( P.X , P.Y ) then HotMap[ N , P.X , P.Y ] := 0;

	{ Increase the timer for this map, since we're dealing with a }
	{ stationary position, so the data should remain good for a }
	{ nice long time. }
	MapUpdate[ N ] := GB^.ComTime + 100;

	HotMapFloodFill( GB , N , MM );
end;

Procedure Calc_EdgeMap( GB: GameBoardPtr; N,PDat,MM: Integer );
	{ Calculate a map seeking the described edges. }
var
	M: GearPtr;
	P: Point;
	X,Y: Integer;
begin
	{ Set the position for all blocking gears on the coldmap. }
	M := GB^.Meks;
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if IsMasterGear( M ) and OnTheMap( P.X , P.Y ) and GearOperational( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end else if ( M^.G = GG_MetaTerrain ) and ( M^.Stat[ STAT_Pass ] <= -100 ) and NotDestroyed( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;

	{ Set the hot spots. }
	for X := 1 to XMax do begin
		if AngDir[ PDat , 2 ] = -1 then HotMap[ N , X , 1 ] := 0
		else if AngDir[ PDat , 2 ] = 1 then HotMap[ N , X , YMax ] := 0;
	end;
	for Y := 1 to YMax do begin
		if AngDir[ PDat , 1 ] = -1 then HotMap[ N , 1, Y ] := 0
		else if AngDir[ PDat , 1 ] = 1 then HotMap[ N , XMax , Y ] := 0;
	end;

	{ Increase the timer for this map, since we're dealing with a }
	{ stationary position, so the data should remain good for a }
	{ nice long time. }
	MapUpdate[ N ] := GB^.ComTime + 100;

	HotMapFloodFill( GB , N , MM );
end;

{$IFDEF PATCH_HEART}
Procedure Calc_NearestEdgeMap( GB: GameBoardPtr; N,MM: Integer );
	{ Calculate a map seeking the nearest edges. }
var
	M: GearPtr;
	P: Point;
	X,Y: Integer;
begin
	{ Set the position for all blocking gears on the coldmap. }
	M := GB^.Meks;
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if IsMasterGear( M ) and OnTheMap( P.X , P.Y ) and GearOperational( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end else if ( M^.G = GG_MetaTerrain ) and ( M^.Stat[ STAT_Pass ] <= -100 ) and NotDestroyed( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;

	{ Set the hot spots. }
	for X := 1 to XMax do begin
		HotMap[ N , X , 1 ] := 0;
		HotMap[ N , X , YMax ] := 0;
	end;
	for Y := 1 to YMax do begin
		HotMap[ N , 1, Y ] := 0;
		HotMap[ N , XMax , Y ] := 0;
	end;

	{ Increase the timer for this map, since we're dealing with a }
	{ stationary position, so the data should remain good for a }
	{ nice long time. }
	MapUpdate[ N ] := GB^.ComTime + 100;

	HotMapFloodFill( GB , N , MM );
end;
{$ENDIF PATCH_HEART}

Procedure Calc_SDMap( GB: GameBoardPtr; N,Team,MM: Integer );
	{ Calculate a SEEK AND DESTROY hotmap for a member of TEAM using }
	{ movement mode MM. }
var
	M: GearPtr;
	Flag: Boolean;
	P: Point;
begin
	{ Set the position for all enemies of the listed team to 0 on hotmap. }
	{ Set the position for all blocking gears on the coldmap. }
	M := GB^.Meks;
	flag := True;
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if IsMasterGear( M ) and OnTheMap( P.X , P.Y ) and GearOperational( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
			if TeamCanSeeTarget( GB , Team , M ) and AreEnemies( GB , Team , NAttValue( M^.NA , NAG_Location , NAS_Team ) ) then begin
				HotMap[ N , P.X , P.Y ] := 0;
				Flag := False;
			end;
		end else if ( M^.G = GG_MetaTerrain ) and ( M^.Stat[ STAT_Pass ] <= -100 ) and NotDestroyed( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;

	{ If no visible enemies found, no point in continuing with the }
	{ algorithm. }
	if Flag then Exit;

	HotMapFloodFill( GB , N , MM );
end;

Procedure Calc_TeamMap( GB: GameBoardPtr; N,Team,MM: Integer );
	{ Calculate a FOLLOW TEAM hotmap for a member of TEAM using }
	{ movement mode MM. }
var
	M: GearPtr;
	Flag: Boolean;
	P: Point;
begin
	{ Set the position for all enemies of the listed team to 0 on hotmap. }
	{ Set the position for all blocking gears on the coldmap. }
	M := GB^.Meks;
	flag := True;
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if IsMasterGear( M ) and OnTheMap( P.X , P.Y ) and GearOperational( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
			if ( NAttValue( M^.NA , NAG_Location , NAS_Team ) = Team ) then begin
				HotMap[ N , P.X , P.Y ] := 0;
				Flag := False;
			end;
		end else if ( M^.G = GG_MetaTerrain ) and ( M^.Stat[ STAT_Pass ] <= -100 ) and NotDestroyed( M ) then begin
			Inc( ColdMap[ N , P.X , P.Y ] );
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;

	{ If no visible enemies found, no point in continuing with the }
	{ algorithm. }
	if Flag then Exit;

	HotMapFloodFill( GB , N , MM );
end;

Procedure CalculateHotMap( GB: GameBoardPtr; N,Team,MM,O: Integer );
	{ Calculate a hot map, choosing type based on the ORDER given. }
var
	X,Y: Integer;
begin
	MapUpdate[n] := GB^.ComTime + Hot_Map_Validity;
	MapTeam[n] := Team;
	MapMoveMode[n] := MM;
	MapOrder[n] := O;

	{ Clear out the entire map, first. }
	For X := 1 to XMax do begin
		For Y := 1 to YMax do begin
{$IFDEF PATCH_GH}
			HotMap[ N , X , Y ] := Hot_Map_UpperLimit;
{$ELSE PATCH_GH}
			HotMap[ N , X , Y ] := 9999;
{$ENDIF PATCH_GH}
			ColdMap[ N , X , Y ] := 0;
		end;
	end;
	UpdateShadowMap( GB );

{$IFDEF PATCH_HEART}
	case O of
	ORD_SeekEnemy:		Calc_SDMap( GB , N , Team , MM );
	ORD_SeekSingleModel:	Calc_FollowMap( GB , N , Team , MM );
	ORD_SeekSpot:		Calc_SpotMap( GB , N , Team , MM );
	ORD_SeekEdge:		Calc_EdgeMap( GB , N , Team , MM );
	ORD_SeekTeam:		Calc_TeamMap( GB , N , Team , MM );
	ORD_SeekNearestEdge:	Calc_NearestEdgeMap( GB , N , MM );
	else			ErrorMessage_fork('ERROR: aibrain.pp: CalculateHotMap(): ORDER is out of range.');
	end;
{$ELSE PATCH_HEART}
	if O = ORD_SeekSingleModel then begin
		Calc_FollowMap( GB , N , Team , MM );
	end else if O = ORD_SeekSpot then begin
		Calc_SpotMap( GB , N , Team , MM );
	end else if O = ORD_SeekEdge then begin
		Calc_EdgeMap( GB , N , Team , MM );
	end else if O = ORD_SeekTeam then begin
		Calc_TeamMap( GB , N , Team , MM );
	end else begin
		Calc_SDMap( GB , N , Team , MM );
	end;
{$ENDIF PATCH_HEART}
end;

Procedure ClearHotMaps;
	{ Clear all the currently processed floodfill maps. }
var
	t: Integer;
begin
	for t := 1 to NumFFMap do MapUpdate[ T ] := -1;
end;

Function GetHotMap( GB: GameBoardPtr; Team,MM,O: Integer ): Integer;
	{ Locate an appropriate HotMap for the data provided. }
var
	RepMap,GoodMap,T: Integer;
begin
{$IFDEF PATCH_GH}
	if (MM < 0) then Exit(0);
{$ENDIF PATCH_GH}

	RepMap := 1;
	GoodMap := 0;
	for T := 1 to NumFFMap do begin
		if ( MapUpdate[ t ] >= GB^.ComTime ) and ( MapTeam[ t ] = Team ) and ( MapMoveMode[ t ] = MM ) and ( MapOrder[ t ] = O ) then begin
			GoodMap := T;
		end;
		if MapUpdate[ t ] < MapUpdate[ RepMap ] then RepMap := T;
	end;

	if GoodMap <> 0 then begin
		GetHotMap := GoodMap;
	end else begin
		CalculateHotMap( GB , RepMap , Team , MM , O );
		GetHotMap := RepMap;
	end;
end;

procedure AIAttacker( GB: GameBoardPtr; Mek,Weapon,Target: GearPtr );
	{ MEK wants to fire WEP at TAR. }
var
	AttSkillVal, DefSkillVal: Integer;	{ Used to calculate the chances of hitting. }
					{ Thanks to Peter Cordes }
	AtOp: Integer;
	T2,T3: GearPtr;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek)    or (Mek^.G <= GG_DisposeGear)    then Exit;
	if (NIL = Weapon) or (Weapon^.G <= GG_DisposeGear) then Exit;
	if (NIL = Target) or (Target^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Calculate AttSkillVal, DefSkillVal, and set AtOp to 0. }
	if Target^.G = GG_Mecha then DefSkillVal := SkillValue( Target, 5 )
	else DefSkillVal := SkillValue( Target, 10 );
	AttSkillVal := SkillValue( Mek , AttackSkillNeeded( Weapon ) ) + CalcTotalModifiers( gb , Weapon , Target , 0 , WeaponAttackAttributes( Weapon ) );
	AtOp := 0;

	{ If the odds of hitting are good enough, the attacker may try }
	{ to make a called shot. }
	if not NoCalledShots(WeaponAttackAttributes(Weapon),0) and (AttSkillVal-DefSkillVal > 4+Random(10))  and (Random(3) <> 1) and ((Weapon^.S <> GS_Ballistic) or (Weapon^.S <> GS_BeamGun) or (Random(6)+Random(6) > Weapon^.Stat[STAT_BurstValue])) then begin
		T2 := Target^.SubCom;

		if WeaponDC( Weapon , 0 ) > 5 then begin
			{ If using a powerful weapon, do aimed shot at the torso. }
			while ( T2 <> Nil ) and (( T2^.G <> GG_Module ) or ( T2^.S <> GS_Body )) do T2 := T2^.Next;

{$IFDEF PATCH_GH}
		end else if (NIL <> T2) and (GG_DisposeGear < T2^.G) then begin
{$ELSE PATCH_GH}
		end else if T2 <> Nil then begin
{$ENDIF PATCH_GH}
			{ If using a less powerful weapon, do aimed shot at the part with the lowest armor. }
			T3 := Nil;
			while T2 <> Nil do begin
{$IFDEF PATCH_GH}
				if (GG_DisposeGear < T2^.G) then begin
					if (NIL = T3) then begin
						T3 := T2
					end else if GearCurrentArmor( T2 ) < GearCurrentArmor( T3 ) then begin
						T3 := T2;
					end;
				end;
{$ELSE PATCH_GH}
				if T3 = Nil then T3 := T2
				else if GearCurrentArmor( T2 ) < GearCurrentArmor( T3 ) then T3 := T2;
{$ENDIF PATCH_GH}
				T2 := T2^.Next;
			end;
			T2 := T3;
		end;

{$IFDEF PATCH_GH}
		if (NIL <> T2) and (GG_DisposeGear < T2^.G) then begin
			Target := T2;
		end;
{$ELSE PATCH_GH}
		if T2 <> Nil then Target := T2;
{$ENDIF PATCH_GH}

	end else if ( Weapon^.G = GG_Weapon ) then begin
		{ If not making a called shot, the attacker will take }
		{ advantage of rapid fire if possible. }
		if ( Weapon^.S = GS_Ballistic ) and ( Weapon^.Stat[ STAT_BurstValue ] > 0 ) then begin
			AtOp := Weapon^.Stat[ STAT_BurstValue ];
		end else if ( Weapon^.S = GS_BeamGun ) and ( Weapon^.Stat[ STAT_BurstValue ] > 0 ) then begin
			AtOp := Weapon^.Stat[ STAT_BurstValue ];
		end else if Weapon^.S = GS_Missile then begin
			AtOp := Random( Weapon^.Stat[ STAT_Magazine ] );
		end;
	end;

	AttackerFrontEnd( GB , Mek , Weapon , Target , AtOp );

	{ Set the initiative recharge counter. }
	SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_InitRecharge , GB^.ComTime + ReactionTime( Mek ) );
end;

{$IFDEF PATCH_CHEAT}
Procedure SelectPurgePartsMode( Mek: GearPtr; GB: GameBoardPtr );
var
	FindParts: Boolean;

	Function CheckParts( P: GearPtr ): Boolean;
	var
		it: Boolean;
		CurrentDamage: Integer;
		MaxDamage: Integer;
		CurrentArmor: Integer;
		MaxArmor: Integer;
	begin
		CurrentDamage := GearCurrentDamage( P );
		MaxDamage     := GearMaxDamage( P );
		CurrentArmor  := GearCurrentArmor(P);
		MaxArmor      := GearMaxArmor(P);
		case P^.G of
			GG_Module:	it := ( CurrentDamage <= ( ( MaxDamage * 90 ) div 100 ) );
			GG_Mecha:	it := False;
			GG_Character:	it := False;
			GG_Cockpit:	it := False;
			GG_Weapon:	it := True;
			GG_Ammo:	it := ( P^.Stat[STAT_AmmoPresent] <= NAttValue( P^.NA , NAG_WeaponModifier , NAS_AmmoSpent ) );
			GG_MoveSys:	it := False;
			GG_Holder:	it := True;
			GG_Sensor:	it := False;
			GG_Support:	it := False;
			GG_Shield:	it := ( CurrentArmor <= ( ( MaxArmor * 90 ) div 100 ) );
			GG_ExArmor:	it := ( CurrentArmor <= ( ( MaxArmor * 90 ) div 100 ) );
			GG_Scene:	it := False;
			GG_Swag:	it := False;
			GG_Prop:	it := False;
			GG_MetaTerrain:	it := False;
			GG_Electronics:	it := False;
			GG_Usable:	it := False;
			GG_RepairFuel:	it := False;
			GG_Consumable:	it := False;
			GG_Modifier:	it := False;
			GG_WeaponAddOn:	it := True;
		else it := True;
		end;
		CheckParts := it;
	end;

	Function CheckPartsAlongTrack( P: GearPtr ): Boolean;
	var
		P_Next: GearPtr;
	begin
		while ( NIL <> P ) do begin
			P_Next := P^.Next;
			if (GG_DisposeGear < P^.G) then begin
				if not CheckParts( P ) then begin
					exit( False );
				end;

				if ( GG_Cockpit = P^.G ) then begin
					{ Don't add parts beyond the cockpit barrier. }
					if ( not CheckPartsAlongTrack( P^.InvCom ) ) then begin
						exit( False );
					end;
				end else begin
					if ( not CheckPartsAlongTrack( P^.SubCom ) ) then begin
						exit( False );
					end;
					if ( not CheckPartsAlongTrack( P^.InvCom ) ) then begin
						exit( False );
					end;
				end;
			end;
			P := P_Next;
		end;
		CheckPartsAlongTrack := True;
	end;

var
	GS_Tag: String;

	Function CheckPartAlongTrack( P: GearPtr ): Boolean;
	var
		P_Next: GearPtr;
	begin
		while ( NIL <> P ) do begin
			P_Next := P^.Next;
			if (GG_DisposeGear < P^.G) then begin
				if ( '' <> SAttValue( P^.SA, GS_Tag ) ) then begin
					FindParts := True;
					if ( not CheckParts( P ) ) then begin
						exit( False );
					end;
					if ( not CheckPartsAlongTrack( P^.SubCom ) ) then begin
						exit( False );
					end;
					if ( not CheckPartsAlongTrack( P^.InvCom ) ) then begin
						exit( False );
					end;
				end else begin
					if ( GG_Cockpit = P^.G ) then begin
						{ Don't add parts beyond the cockpit barrier. }
						if ( not CheckPartAlongTrack( P^.InvCom ) ) then begin
							exit( False );
						end;
					end else begin
						if ( not CheckPartAlongTrack( P^.SubCom ) ) then begin
							exit( False );
						end;
						if ( not CheckPartAlongTrack( P^.InvCom ) ) then begin
							exit( False );
						end;
					end;
				end;
			end;
			P := P_Next;
		end;
		CheckPartAlongTrack := True;
	end;

var
	MaxMode: Integer;
	T: Integer;
	SelectMode: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then exit;
{$ENDIF PATCH_GH}

	MaxMode := SAttValueToInt( Mek^.SA, SATT_SEPARABLE );
	if ( MaxMode < 1 ) then exit;

	SelectMode := 0;
	for T := 1 to MaxMode do begin
		if ( 0 <= SAttValueToInt(Mek^.SA,SATT_SEPARATE_WAIT + BStr(T)) ) then begin
			FindParts := False;
			GS_Tag := SATT_SEPARATE + BStr(T);
			if CheckPartAlongTrack( Mek^.SubCom ) then begin
				if FindParts then begin
					SelectMode := T;
				end;
			end;
		end;
	end;

	if ( SelectMode < 1 ) or ( MaxMode < SelectMode ) then begin
		exit;
	end;

	DoPurgeParts( GB , Mek , SelectMode );
end;
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_CHEAT}
Procedure SelectFastestFormMode( Mek: GearPtr; GB: GameBoardPtr );
var
	MaxForm: Integer;
	OldForm: Integer;
	MaxSpeed: Integer;
	FM: Integer;
	FS: Integer;
	WaitTime: Integer;
	I: Integer;
	J: Integer;
	NewForm: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	MaxForm := SAttValueToInt( Mek^.SA , SATT_TRANSFORMABLE );
	if MaxForm < 1 then exit;
	OldForm := SAttValueToInt( Mek^.SA , SATT_TRANSFORM_CURRENT );
	if OldForm < 1 then exit;
	if not CheckConversionSystem( Mek , False ) then begin
		exit;
	end;

	MaxSpeed := 0;
	NewForm := 0;

	for I := 1 to MaxForm do begin
		WaitTime := SAttValueToInt( Mek^.SA , SATT_TRANSFORM_WAIT + BStr(OldForm) + '_' + BStr(I) );
		if ( ( 0 < WaitTime ) or ( OldForm = I ) ) then begin
			FM := SAttValueToInt( Mek^.SA , SATT_TRANSFORM_GS + BStr(I) );
			for J := 1 to NumMoveMode do begin
				FS := FormSpeedLimit[ FM , J ];
				if ( MaxSpeed < FS ) then begin
					MaxSpeed := FS;
					NewForm := I;
				end else if ( MaxSpeed = FS ) then begin
					if ( 1 = random(2) ) then begin
						MaxSpeed := FS;
						NewForm := I;
					end;
				end;
			end;
		end;
	end;

	if ( ( 0 < NewForm ) and ( OldForm <> NewForm ) ) then begin
		DoTransformation( GB , Mek , NewForm );
	end;
end;

Procedure SelectHeavyFormMode( Mek: GearPtr; GB: GameBoardPtr );
const
	FormVolume: Array [0..NumForm-1] of Integer = (
		50,	{Battroid}
		20,	{Zoanoid}
		20,	{GroundHugger}
		20,	{Arachnoid}
		20,	{AeroFighter}
		20,	{Ornithoid}
		50,	{GerWalk}
		20,	{HoverFighter}
		20	{GroundCar}
	);
var
	MaxForm: Integer;
	OldForm: Integer;
	MaxVolume: Integer;
	FM: Integer;
	FV: Integer;
	WaitTime: Integer;
	I: Integer;
	NewForm: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	MaxForm := SAttValueToInt( Mek^.SA , SATT_TRANSFORMABLE );
	if MaxForm < 1 then exit;
	OldForm := SAttValueToInt( Mek^.SA , SATT_TRANSFORM_CURRENT );
	if OldForm < 1 then exit;
	if not CheckConversionSystem( Mek , False ) then begin
		exit;
	end;

	MaxVolume := 0;
	NewForm := 0;

	for I := 1 to MaxForm do begin
		WaitTime := SAttValueToInt( Mek^.SA , SATT_TRANSFORM_WAIT + BStr(OldForm) + '_' + BStr(I) );
		if ( ( 0 < WaitTime ) or ( OldForm = I ) ) then begin
			FM := SAttValueToInt( Mek^.SA , SATT_TRANSFORM_GS + BStr(I) );
			FV := FormVolume[ FM ];
			if ( random(100) < FV ) then begin
				NewForm := I;
				break;
			end;
		end;
	end;

	if ( ( 0 < NewForm ) and ( OldForm <> NewForm ) ) then begin
		DoTransformation( GB , Mek , NewForm );
	end;
end;
{$ENDIF PATCH_CHEAT}

procedure SelectMoveMode( Mek: GearPtr; GB: GameBoardPtr );
	{ Set the mek's MoveMode attribute to the highest }
	{ active movemode that this mek has. }
var
	T,MM,MaxSpeed: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	MM := 0;
	MaxSpeed := 0;
	for T := 1 to NumMoveMode do begin
		if ( BaseMoveRate( Mek , T ) > MaxSpeed ) and MoveLegal( Mek , T , NAV_NormSpeed , GB^.Comtime ) then begin
			if not ( ( GB <> Nil ) and ( GB^.Scale > Mek^.Scale ) and ( T = MM_Fly ) and ( JumpTime( Mek ) > 0 ) ) then begin
				MM := T;
				MaxSpeed := BaseMoveRate( Mek , T );
			end;
		end;
	end;
	if MM <> 0 then SetNAtt( Mek^.NA , NAG_Action , NAS_MoveMode , MM);
end;

{$IFDEF PATCH_HEART}
function SafeToFire( GB: GameBoardPtr; Mek, Part, Target: GearPtr ): Boolean;
	{ Return TRUE if firing this weapon won't cause a huge problem, }
	{ or FALSE otherwise. }
var
	Found: Boolean;
	R: Integer;
	P: Point;
	M2: GearPtr;
begin
	{ First off, dumb things won't care if it's safe to fire or not. }
	if ( Mek^.G = GG_Character ) and ( Mek^.Stat[ STAT_Knowledge ] < RollStep( 5 ) ) then begin
		Exit( True );
	end;

	{ Check to see if this is a blast weapon. }
	if HasAttackAttribute( WeaponAttackAttributes( Part ) , AA_BlastAttack ) then begin
		{ If there's an ally within the blast radius, this weapon isn't safe. }
		P := GearCurrentLocation( Target );
		M2 := GB^.Meks;
		R := BlastRadius( GB , Part , WeaponAttackAttributes( Part ) );
		Found := False;
		while M2 <> Nil do begin
{$IFDEF PATCH_GH}
			if (GG_DisposeGear < M2^.G) then begin
{$ENDIF PATCH_GH}
			if AreAllies( GB , Mek , M2 ) and ( Range( M2 , P.X , P.Y ) < R ) then Found := True;
{$IFDEF PATCH_GH}
			end;
{$ENDIF PATCH_GH}
			M2 := M2^.Next;
		end;
		SafeToFire := Not Found;
	end else begin
		{ The blastattack check is the only one I'm doing yet. }
		SafeToFire := True;
	end;
end;

function WGoodness( GB: GameBoardPtr; Part, Target: GearPtr ): Integer;
	{ This is the heuristic SeekBigWeapon uses }
var
	WG,ASV,AM		 : Integer;
	AttSkillVal, DefSkillVal : Integer;
begin
	{ Can your lancemates size up the abilities of a defender in a }
	{ mecha, very far away, in heavy cover?  Sure they can! }
	{ Just look at Piloting or Dodge, not talents, parrying, ... }
	if Target^.G = GG_Mecha then DefSkillVal := SkillValue( Target, 5 )
	else DefSkillVal := SkillValue( Target, 10 );

	{ NOTE: the actual attack code does RollStep(SkillValue) + }
	{ modifiers, so this is an approximation }
	{ Missiles will have BustValue = 0.  Don't fire }
	{ until you can see the whites of their eyes. }
	ASV := SkillValue( FindRoot(Part) , AttackSkillNeeded( Part ) );
	AM := CalcTotalModifiers( GB , Part , Target , Part^.Stat[ STAT_BurstValue ] , WeaponAttackAttributes( Part ) );
	AttSkillVal := ASV + AM;

	WG := Part^.V * (1+Part^.Stat[ STAT_BurstValue ]) + 2*(AttSkillVal-DefSkillVal);
	for ASV := 1 to Part^.Scale do WG := WG * 3;
	WGoodness := WG;
end;
{$ENDIF PATCH_HEART}
Function SelectBestWeapon( GB: GameBoardPtr; Mek,Target: GearPtr ): GearPtr;
	{ Select the best weapon for attacking TARGET with. }
var
	weapon: GearPtr;
	BestWeight: LongInt;
{$IFDEF PATCH_HEART}
{$ELSE PATCH_HEART}
	function SafeToFire( Part: GearPtr ): Boolean;
		{ Return TRUE if firing this weapon won't cause a huge problem, }
		{ or FALSE otherwise. }
	var
		Found: Boolean;
		R: Integer;
		P: Point;
		M2: GearPtr;
	begin
		{ First off, dumb things won't care if it's safe to fire or not. }
		if ( Mek^.G = GG_Character ) and ( Mek^.Stat[ STAT_Knowledge ] < RollStep( 5 ) ) then begin
			Exit( True );
		end;

		{ Check to see if this is a blast weapon. }
		if HasAttackAttribute( WeaponAttackAttributes( Part ) , AA_BlastAttack ) then begin
			{ If there's an ally within the blast radius, this weapon isn't safe. }
			P := GearCurrentLocation( Target );
			M2 := GB^.Meks;
			R := BlastRadius( GB , Part , WeaponAttackAttributes( Part ) );
			Found := False;
			while M2 <> Nil do begin
{$IFDEF PATCH_GH}
				if (GG_DisposeGear < M2^.G) then begin
{$ENDIF PATCH_GH}
				if AreAllies( GB , Mek , M2 ) and ( Range( M2 , P.X , P.Y ) < R ) then Found := True;
{$IFDEF PATCH_CHEAT}
				if Cheat_NPC_AvoidAttackToNeutrality and ( not AreEnemies( GB , Mek , M2 ) ) and ( Range( M2 , P.X , P.Y ) < R ) then begin
					Found := True;
				end;
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_GH}
				end;
{$ENDIF PATCH_GH}
				M2 := M2^.Next;
			end;
			SafeToFire := Not Found;
		end else begin
			{ The blastattack check is the only one I'm doing yet. }
			SafeToFire := True;
		end;
	end;

	function WGoodness( Part, Target: GearPtr ): Integer;
		{ This is the heuristic SeekBigWeapon uses }
	var
		WG,ASV,AM		 : Integer;
		AttSkillVal, DefSkillVal : Integer;
	begin
		{ Can your lancemates size up the abilities of a defender in a }
		{ mecha, very far away, in heavy cover?  Sure they can! }
		{ Just look at Piloting or Dodge, not talents, parrying, ... }
		if Target^.G = GG_Mecha then DefSkillVal := SkillValue( Target, 5 )
		else DefSkillVal := SkillValue( Target, 10 );

		{ NOTE: the actual attack code does RollStep(SkillValue) + }
		{ modifiers, so this is an approximation }
		{ Missiles will have BustValue = 0.  Don't fire }
		{ until you can see the whites of their eyes. }
		ASV := SkillValue( FindRoot(Part) , AttackSkillNeeded( Part ) );
		AM := CalcTotalModifiers( gb , Part , Target , Part^.Stat[ STAT_BurstValue ] , WeaponAttackAttributes( Part ) );
		AttSkillVal := ASV + AM;

		WG := Part^.V * (1+Part^.Stat[ STAT_BurstValue ]) + 2*(AttSkillVal-DefSkillVal);
		for ASV := 1 to Part^.Scale do WG := WG * 3;
		WGoodness := WG;
	end;
{$ENDIF PATCH_HEART}

	procedure SeekBigWeapon( Part: GearPtr );
		{ Seek a weapon which is capable of hitting target. }
		{ Select the best weapon, weighted by accuracy and damage }
		{ Things we don't do: }
		{   * scale factors }
		{   * avoid wasting ammo/recharge time on easy-to-kill targets }
		{	(it would be worth working on the target selection }
		{	algorithm if you were going to do anything about this. }
		{	Maybe find your best weapon and look for targets to }
		{	fire it at.  This could help a lot with LINE weapons. }
		{	Or even find a weight for every weapon/target pair...) }
	var
		{Range : Integer;}
		Weight : Integer;
	begin
		while ( Part <> Nil ) do begin
{$IFDEF PATCH_CHEAT}
			if ( Part^.G = GG_Module ) or ( Part^.G = GG_Weapon ) or ( Cheat_NPC_UseGrenade and ( GG_Ammo = Part^.G ) ) then begin
{$ELSE PATCH_CHEAT}
			if ( Part^.G = GG_Module ) or ( Part^.G = GG_Weapon ) then begin
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_GH}
				if NeedAmmo( Part ) then begin
					ReloadAmmo( GB , Mek , Part );
				end;
{$ENDIF PATCH_GH}
{$IFDEF PATCH_HEART}
				if ReadyToFire( GB , Mek , Part ) and RangeArcCheck( GB , Mek , Part , Target ) and SafeToFire( GB , Mek , Part , Target ) then
{$ELSE PATCH_HEART}
				if ReadyToFire( GB , Mek , Part ) and RangeArcCheck( GB , Mek , Part , Target ) and SafeToFire( Part ) then
{$ENDIF PATCH_HEART}
				begin

					if Weapon = Nil then begin
						Weapon := Part;
{$IFDEF PATCH_HEART}
						BestWeight := WGoodness(GB, Part, Target);
{$ELSE PATCH_HEART}
						BestWeight := WGoodness(Part, Target);
{$ENDIF PATCH_HEART}
					end else begin
{$IFDEF PATCH_HEART}
						Weight := WGoodness(GB, Part, Target);
{$ELSE PATCH_HEART}
						Weight := WGoodness(Part, Target);
{$ENDIF PATCH_HEART}
						{  BestWeight is a variable in the parent procedure.  I don't know how else to do this in Pascal... }
						if Weight > BestWeight then begin
							Weapon := Part;
							BestWeight := Weight;
						end;
					end;
				end;
			end;
			if ( Part^.SubCom <> Nil ) then SeekBigWeapon( Part^.SubCom );
			if ( Part^.InvCom <> Nil ) then SeekBigWeapon( Part^.InvCom );
			Part := Part^.Next;
		end;
	end;

begin
{$IFDEF PATCH_GH}
	if (NIL = Mek)    or (Mek^.G <= GG_DisposeGear)    then Exit(NIL);
	if (NIL = Target) or (Target^.G <= GG_DisposeGear) then Exit(NIL);
{$ENDIF PATCH_GH}
	Weapon := Nil;
	BestWeight := -10000;
	SeekBigWeapon( Mek^.SubCom );
	SeekBigWeapon( Mek^.InvCom );
	SelectBestWeapon := Weapon;
end;

Procedure AttackTargetOfOppurtunity( GB: GameBoardPtr; Mek: GearPtr );
	{ Look for the most oppurtune target to attack. }
var
	weapon: GearPtr;
	TL,Target: GearPtr;
	BestWeight: Longint;
{ *** PROCEDURES BLOCK *** }
	procedure SeekFarWeapon( Part: GearPtr );
		{ Find the weapon with the longest current range. }
	begin
		while ( Part <> Nil ) do begin
{$IFDEF PATCH_CHEAT}
			if ( Part^.G = GG_Module ) or ( Part^.G = GG_Weapon ) or ( Cheat_NPC_UseGrenade and ( GG_Ammo = Part^.G ) ) then begin
{$ELSE PATCH_CHEAT}
			if ( Part^.G = GG_Module ) or ( Part^.G = GG_Weapon ) then begin
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_GH}
				if NeedAmmo( Part ) then begin
					ReloadAmmo( GB , Mek , Part );
				end;
{$ENDIF PATCH_GH}
				if ReadyToFire( GB , Mek , Part ) then begin
					if Weapon = Nil then Weapon := Part
					else if WeaponRange( GB , Part ) > WeaponRange( GB , Weapon ) then Weapon := Part;
				end;
			end;
			if ( Part^.SubCom <> Nil ) then SeekFarWeapon( Part^.SubCom );
			if ( Part^.InvCom <> Nil ) then SeekFarWeapon( Part^.InvCom );
			Part := Part^.Next;
		end;
	end;

	function SafeToFire( Part: GearPtr ): Boolean;
		{ Return TRUE if firing this weapon won't cause a huge problem, }
		{ or FALSE otherwise. }
	var
		Found: Boolean;
		N, R, T, TT: Integer;
		P, P1, P2: Point;
		M2: GearPtr;
	begin
		{ First off, dumb things won't care if it's safe to fire or not. }
		if ( Mek^.G = GG_Character ) and ( Mek^.Stat[ STAT_Knowledge ] < Random( 6 ) ) then begin
			Exit( True );
		end;

		Found := False;

		{ Check to see if this is a blast weapon. }
		if HasAttackAttribute( WeaponAttackAttributes( Part ) , AA_BlastAttack ) then begin
			{ If there's an ally within the blast radius, this weapon isn't safe. }
			P := GearCurrentLocation( Target );
			M2 := GB^.Meks;
			R := BlastRadius( GB , Part , WeaponAttackAttributes( Part ) );

			while M2 <> Nil do begin
{$IFDEF PATCH_GH}
				if (GG_DisposeGear < M2^.G) then begin
{$ENDIF PATCH_GH}
				if AreAllies( GB , Mek , M2 ) and ( Range( M2 , P.X , P.Y ) < R ) then Found := True;
{$IFDEF PATCH_CHEAT}
				if Cheat_NPC_AvoidAttackToNeutrality and ( not AreEnemies( GB , Mek , M2 ) ) and ( Range( M2 , P.X , P.Y ) < R ) then begin
					Found := True;
				end;
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_GH}
				end;
{$ENDIF PATCH_GH}
				M2 := M2^.Next;
			end;

		end else if HasAttackAttribute( WeaponAttackAttributes( Part ) , AA_LineAttack ) then begin
			{ code from effects.pp:DoLineAttack() }
			P1 := GearCurrentLocation( Part );
			P1.Z := MekAltitude( GB , Part );
			P2 := GearCurrentLocation( Target );
			P2.Z := MekAltitude( GB, Target );

			P.Z := P1.Z;
{$IFDEF PATCH_GH}
			R := SolveLine_Get_RangedN( P1.X , P1.Y , P1.Z , P2.X , P2.Y , P2.Z , WeaponRange( GB, Part ) );
{$ELSE PATCH_GH}
			R := WeaponRange( GB, Part );
{$ENDIF PATCH_GH}
			T := 0;
			UpdateShadowMap( GB );
			while ( T < R ) do begin
				Inc( T );
				P := SolveLine( P1.X , P1.Y , P1.Z , P2.X , P2.Y , P2.Z , T );
				if OnTheMap( P.X , P.Y ) then begin
					N := NumGearsXY( GB , P.X , P.Y );
					if N > 0 then begin
						for tt := 1 to N do begin
							M2 := FindGearXY( GB , P.X , P.Y , TT );
							if (Abs(MekAltitude( GB , M2 ) - P.Z ) <= 1) and NotDestroyed( M2 ) and AreAllies( GB , Mek , M2 ) then Found := True;
{$IFDEF PATCH_CHEAT}
							if Cheat_NPC_AvoidAttackToNeutrality and (Abs(MekAltitude( GB , M2 ) - P.Z ) <= 1) and NotDestroyed( M2 ) and ( not AreEnemies( GB , Mek , M2 ) ) then begin
								Found := True;
							end;
{$ENDIF PATCH_CHEAT}
						end;
					end;

					if TileBlocksLOS( GB , P.X , P.Y , P.Z ) then T := R;
				end;
			end;

		end;
		{ The blastattack check is the only one I'm doing yet. }
		{ Line attack added.  Nothing sucks like your non-sentient robot }
		{ killing you with a plasma cannon in one shot }

		SafeToFire := Not Found;
	end;


begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ First, check to make sure that the mecha hasn't attacked }
	{ too recently. }
	if NAttValue( Mek^.NA , NAG_EpisodeData , NAS_InitRecharge ) > GB^.ComTime then Exit;

	{ Start by finding a good weapon to fire with. }
	{ Preference will be given to the weapon with the longest range. }
	Weapon := Nil;
	BestWeight := -10000;
	SeekFarWeapon( Mek^.SubCom );
	SeekFarWeapon( Mek^.InvCom );

	if Weapon <> Nil then begin
		{ Next, see if we have an appropriate target. }
		Target := Nil;
		TL := gb^.meks;

		while TL <> Nil do begin
{$IFDEF PATCH_GH}
			if (GG_DisposeGear < TL^.G) then begin
				if (AreEnemies( GB , Mek , TL ) and OnTheMap( TL )) then begin
					if  (RangeArcCheck( GB , Mek , Weapon , TL ) and GearActive( TL )) then begin
						if MekCanSeeTarget( GB , Mek , TL ) then begin
							if Target = Nil then begin
								Target := TL
							end else begin
								if (Range( gb , Target , Mek ) > Range( gb , TL , Mek )) then begin
									Target := TL;
								end;
							end;
						end;
					end;
				end;
			end;
{$ELSE PATCH_GH}
			if AreEnemies( GB , Mek , TL ) and OnTheMap( TL ) and RangeArcCheck( GB , Mek , Weapon , TL ) and GearActive( TL ) and MekCanSeeTarget( GB , Mek , TL ) then begin
				if Target = Nil then Target := TL
				else if Range( gb , Target , Mek ) > Range( gb , TL , Mek ) then Target := TL;
			end;
{$ENDIF PATCH_GH}
			TL := TL^.Next;
		end;

		{ If a target was found, fire at that target with the }
		{ biggest weapon in the arsenal. }
		if Target <> Nil then begin
			Weapon := SelectBestWeapon( GB , Mek , Target );
			if Weapon <> Nil then AIAttacker( GB , Mek , Weapon , Target );
		end;
	end;
end;

Procedure Wander( Mek: GearPtr; GB: GameBoardPtr );
	{ Just wander about. This procedure is often called when }
	{ conventional pathfinding methods have failed. It is also }
	{ used as the "SEEK" part of "SEEK AND DESTROY". }
var
	P: Point;
	CD: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Determine current facing and position. }
	P := GearCurrentLocation( Mek );
	CD := NAttValue( Mek^.NA , NAG_Location , NAS_D );

	{ If the current direction of travel doesn't lead off the map, }
	{ and isn't blocked, just move foreword. }
	if OnTheMap( P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ) and ( Random( 5 ) <> 1 ) and not MoveBlocked( Mek , GB ) then begin
		if Mek^.G = GG_Mecha then begin
			PrepAction( GB , Mek , NAV_FullSpeed );
		end else begin
			PrepAction( GB , Mek , NAV_NormSpeed );
		end;
	end else begin
		if Random( 2 ) = 1 then begin
			PrepAction( GB , Mek , NAV_TurnRight );
		end else begin
			PrepAction( GB , Mek , NAV_TurnLeft );
		end;
	end;
end;

Function XMoveTowardsGoal( GB: GameBoardPtr; Mek: GearPtr; HM,OptMax,OptMin: Integer ): Boolean;
	{ Using hotmap N as a guide, attempt to move towards the goal. }
	{ OPTMAX and OPTMIN describe the maximum and minumum hotmap values the mek will }
	{ try to reach. If its current location has a hot value lower than OptMax, it }
	{ will stay in that spot. If its current location is lower than OptMin, it will }
	{ attempt to move to a location with a higher hot value. }
	Function ThisMoveIsOkay( MAct,X,Y: Integer ): Boolean;
		{ This function combines two checks: Whether or not the target tile is }
		{ unblocked, and whether or not the specified moveaction is legal. }
	begin
		ThisMoveIsOkay := MoveLegal( Mek , MAct , GB^.ComTime ) and not IsBlocked( Mek , GB , X , Y );
	end;
var
	P,P2: Point;
	T,D,Best,CD: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit(False);
{$ENDIF PATCH_GH}

	P := GearCurrentLocation( Mek );

	{ If our current direction of travel is bringing us closer to the }
	{ target, keep moving in that direction, even if it isn't the }
	{ optimal path. }
	CD := NAttValue( Mek^.NA , NAG_Location , NAS_D );
	if HotMap[ HM , P.X ,P.Y ] < OptMin then begin
		{ Too close to the target- move farther away. }
		P2.X := P.X + AngDir[ ( CD + 4 ) mod 8 , 1 ];
		P2.Y := P.Y + AngDir[ ( CD + 4 ) mod 8 , 2 ];
		if OnTheMap( P2.X , P2.Y ) and ( HotMap[ HM , P2.X , P2.Y ] > HotMap[ HM , P.X , P.Y ] ) and ThisMoveIsOkay( NAV_Reverse , P2.X , P2.Y ) then begin
			PrepAction( GB , Mek , NAV_Reverse );
			XMoveTowardsGoal := True;

		end else if OnTheMap( P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ) and ( HotMap[ HM , P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ] > HotMap[ HM , P.X , P.Y ] ) and ThisMoveIsOkay( NAV_NormSpeed , P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ) then begin
			if Mek^.G = GG_Character then begin
				PrepAction( GB , Mek , NAV_NormSpeed );
			end else begin
				PrepAction( GB , Mek , NAV_FullSpeed );
			end;
			XMoveTowardsGoal := True;

		end else begin
			D := -1;
			Best := HotMap[ HM , P.X , P.Y ];
			for t := 0 to 7 do begin
				if OnTheMap( P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ) then begin
					if ( HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ] > Best ) and ( ColdMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ] < 1 ) and not IsBlocked( Mek , GB , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ) then begin
						Best := HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ];
						D := T;
					end;
				end;
			end;

			if D <> -1 then begin
				if CD <> D then begin
					Best := NAV_TurnRight;
					for t := 1 to 4 do begin
						if (( CD + T ) mod 8 ) = D then Best := NAV_TurnRight
						else if (( CD + 8 - T ) mod 8 ) = D then Best := NAV_TurnLeft;
					end;
					PrepAction( GB , Mek , Best );

				end else begin
					PrepAction( GB , Mek , NAV_FullSpeed );
				end;
				XMoveTowardsGoal := True;

			end else begin
				{ If we don't have a good place to go, }
				{ try changing move modes to the lowest. }
				if Random( 2 ) = 1 then begin
					GearUp( Mek );
				end;
				XMoveTowardsGoal := False;
			end;
		end;

	end else if ( HotMap[ HM , P.X ,P.Y ] < OptMax ) and ( ( Random( 15 ) < HotMap[ HM , P.X ,P.Y ] ) or IsInCover( GB , Mek ) ) then begin
		{ Within optimal range. Turn to hopefully face a target. }
		D := -1;
		Best := HotMap[ HM , P.X , P.Y ];
		for t := 0 to 7 do begin
			if OnTheMap( P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ) then begin
				if ( HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ] < Best ) then begin
					Best := HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ];
					D := T;
				end;
			end;
		end;

		{ If a good direction was found, turn to face it. }
		{ Otherwise assume that we're already facing a good direction. }
		if D <> -1 then begin
			if CD <> D then begin
				Best := NAV_TurnRight;
				for t := 1 to 4 do begin
					if (( CD + T ) mod 8 ) = D then Best := NAV_TurnRight
					else if (( CD + 8 - T ) mod 8 ) = D then Best := NAV_TurnLeft;
				end;
				PrepAction( GB , Mek , Best );
			end else begin
				WaitAMinute( GB , Mek , ReactionTime( Mek ) );
			end;
		end else begin
			{ If we don't have a good place to go, }
			{ try changing move modes to the lowest. }
			if Random( 2 ) = 1 then begin
				GearUp( Mek );
			end;
			WaitAMinute( GB , Mek , ReactionTime( Mek ) );
		end;
		XMoveTowardsGoal := True;


	end else if OnTheMap( P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ) and ( HotMap[ HM , P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ] < HotMap[ HM , P.X , P.Y ] ) and ( HotMap[ HM , P.X , P.Y ] < Random( 4 ) ) and not MoveBlocked( Mek , GB ) then begin
{$IFDEF PATCH_CHEAT}
		if (( Mek^.G = GG_Mecha ) or ( not Cheat_DoNotRunWhenMouseClicked and ( CurrentStamina( Mek ) > 10 ) ) ) and MoveLegal( Mek , NAV_FullSpeed , GB^.ComTime ) then begin
{$ELSE PATCH_CHEAT}
		if (( Mek^.G = GG_Mecha ) or ( CurrentStamina( Mek ) > 10 ) ) and MoveLegal( Mek , NAV_FullSpeed , GB^.ComTime ) then begin
{$ENDIF PATCH_CHEAT}
			PrepAction( GB , Mek , NAV_FullSpeed );
		end else if MoveLegal( Mek , NAV_NormSpeed , GB^.ComTime ) then begin
			PrepAction( GB , Mek , NAV_NormSpeed );
		end else begin
			Exit( False );
		end;
		XMoveTowardsGoal := True;

	end else begin
		{ Check to see if there's a better direction we can be moving in. }
		D := -1;
		Best := HotMap[ HM , P.X , P.Y ];
		for t := 0 to 7 do begin
			if OnTheMap( P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ) then begin
				if ( HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ] < Best ) and ( ColdMap[ HM , P.X + AngDir[T,1] , P.Y + AngDir[T,2] ] < 1 ) and not IsBlocked( Mek,GB,P.X + AngDir[T,1] , P.Y + AngDir[T,2]) then begin
					Best := HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ];
					D := T;
				end;
			end;
		end;

		if D <> -1 then begin
			if CD <> D then begin
				Best := NAV_TurnRight;
				for t := 1 to 4 do begin
					if (( CD + T ) mod 8 ) = D then Best := NAV_TurnRight
					else if (( CD + 8 - T ) mod 8 ) = D then Best := NAV_TurnLeft;
				end;
				PrepAction( GB , Mek , Best );

			end else begin
				if MoveBlocked( Mek , GB ) then begin
					Exit( False );
{$IFDEF PATCH_CHEAT}
				end else if ( HotMap[ HM , P.X ,P.Y ] > ( OptMax * 3 div 2 ) ) and (( Mek^.G = GG_Mecha ) or ( not Cheat_DoNotRunWhenMouseClicked and ( CurrentStamina( Mek ) > 10 ) ) ) and MoveLegal( Mek , NAV_FullSpeed, GB^.ComTime ) then begin
{$ELSE PATCH_CHEAT}
				end else if ( HotMap[ HM , P.X ,P.Y ] > ( OptMax * 3 div 2 ) ) and (( Mek^.G = GG_Mecha ) or ( CurrentStamina( Mek ) > 10 ) ) and MoveLegal( Mek , NAV_FullSpeed, GB^.ComTime ) then begin
{$ENDIF PATCH_CHEAT}
					PrepAction( GB , Mek , NAV_FullSpeed );
				end else if MoveLegal( Mek , NAV_NormSpeed, GB^.ComTime ) then begin
					PrepAction( GB , Mek , NAV_NormSpeed );
				end else begin
					Exit( False );
				end;
			end;
			XMoveTowardsGoal := True;

		end else begin
			{ If we don't have a good place to go, }
			{ try changing move modes to the lowest. }
			if Random( 2 ) = 1 then begin
				GearUp( Mek );
			end;
			XMoveTowardsGoal := False;
		end;
	end;
end;

Procedure MoveTowardsGoal( GB: GameBoardPtr; Mek: GearPtr; HM: Integer );
	{ Front-end for the Extended Move Towards Goal. }
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	if not XMoveTowardsGoal( GB , Mek , HM , 0 , 0 ) then Wander( Mek , GB );
end;


Function DoorPresent( GB: GameBoardPtr; X,Y: Integer ): Boolean;
	{ Return TRUE if there's a door here which a passive NPC shouldn't pass. }
begin
	if not OnTheMap( X , Y ) then begin
		DoorPresent := False;
	end else begin
		DoorPresent := GB^.Map[ X , Y ].terr = TERRAIN_Threshold;
	end;
end;

Procedure MillAround( GB: GameBoardPtr; Mek: GearPtr );
	{ Stay in roughly the same position, wander around a little bit. }
var
	P: Point;
	D: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ No reason to panic. Just stand around; }
	{ maybe move if it's okay. }
	if Random( 3 ) = 1 then begin
		PrepAction( GB , Mek , NAV_Stop );
	end else if Random( 3 ) = 1 then begin
		{ The passive character may walk sometimes. }
		P := GearCurrentLocation( Mek );
		D := NAttValue( Mek^.NA , NAG_Location , NAS_D );
		P.X := P.X + AngDir[ D , 1 ];
		P.Y := P.Y + AngDir[ D , 2 ];
		if OnTheMap( P.X , P.Y ) then begin
			{ Don't move foreword if the move is blocked }
			{ or if the move would take the character over }
			{ the threshold of a door. }
			if IsBlocked( Mek , GB , P.X , P.Y ) or DoorPresent( GB , P.X , P.Y ) then begin
				PrepAction( GB , Mek , NAV_Stop );
			end else begin
				PrepAction( GB , Mek , NAV_NormSpeed );
			end;
		end;
	end else begin
		if Random( 2 ) = 1 then begin
			PrepAction( GB , Mek , NAV_TurnRight );
		end else begin
			PrepAction( GB , Mek , NAV_TurnLeft );
		end;
	end;
end;

Procedure FleeFromGoal( GB: GameBoardPtr; Mek: GearPtr; HM: Integer );
	{ Using hotmap N as a guide, attempt to move away from the goal. }
var
	P: Point;
	T,D,Best,CD: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	P := GearCurrentLocation( Mek );

	{ If our current direction of travel is bringing us closer to the }
	{ target, keep moving in that direction, even if it isn't the }
	{ optimal path. }
	CD := NAttValue( Mek^.NA , NAG_Location , NAS_D );
	if OnTheMap( P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ) and ( HotMap[ HM , P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ] > HotMap[ HM , P.X , P.Y ] ) and ( HotMap[ HM , P.X , P.Y ] > Random( 8 ) ) and not MoveBlocked( Mek , GB ) then begin
		PrepAction( GB , Mek , NAV_NormSpeed );

	end else begin
		D := -1;
		Best := HotMap[ HM , P.X , P.Y ];
		for t := 0 to 7 do begin
			if OnTheMap( P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ) then begin
				if ( HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ] > Best ) and ( ColdMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ] < 1 ) then begin
					Best := HotMap[ HM , P.X + AngDir[ T , 1 ] , P.Y + AngDir[ T , 2 ] ];
					D := T;
				end;
			end;
		end;



		if D <> -1 then begin
			if CD <> D then begin
				Best := NAV_TurnRight;
				for t := 1 to 4 do begin
					if (( CD + T ) mod 8 ) = D then Best := NAV_TurnRight
					else if (( CD + 8 - T ) mod 8 ) = D then Best := NAV_TurnLeft;
				end;
				PrepAction( GB , Mek , Best );

			end else begin
				PrepAction( GB , Mek , NAV_NormSpeed );
			end;
		end else begin
			MillAround( GB , Mek );
		end;
	end;

end;

Function HotMoveMode( Mek: GearPtr ): Integer;
	{ Return the current move mode of this mecha, adjusted for }
	{ characters who are walking. }
var
	T: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit(-1);
{$ENDIF PATCH_GH}

	T := NAttValue( Mek^.NA , NAG_Action , NAS_MoveMode );
	if ( T = MM_Walk ) and ( Mek^.G = GG_Character ) then T := 0;
	HotMoveMode := T;
end;

{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
Function AIRepair( GB: GameBoardPtr; NPC,Target: GearPtr; RepairLimit: LongInt; Skill: Integer ): LongInt;
{$ELSE}
Procedure AIRepair( GB: GameBoardPtr; NPC,Target,RepairFuel: GearPtr; Skill: Integer );
{$ENDIF}
	{ This procedure acts as a frontend for the repair skill bits. }
	{ It's analogous to the DoFieldRepair skill in the backpack unit. }
var
	N: LongInt;
	msg: String;
{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
	NPC_Mek, NPC_Pilot: GearPtr;
	MekRepairFuel, PilotRepairFuel, TotalRepairFuel: LongInt;
{$ENDIF PATCH_GH}
begin
{$IFDEF PATCH_GH}
	if (NIL = NPC) or (NPC^.G <= GG_DisposeGear) then Exit;
	if (NIL = Target) or (Target^.G <= GG_DisposeGear) then Exit;
	{ if (NIL = RepairFuel) or (RepairFuel^.G <= GG_DisposeGear) then Exit; }
{$ENDIF PATCH_GH}

{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
	NPC_Mek := FindRoot( NPC );
	NPC_Pilot := LocatePilot( NPC );
	MekRepairFuel := 0;
	PilotRepairFuel := 0;
	if ( ( NIL <> NPC_Mek ) and ( NPC_Mek <> NPC_Pilot ) ) then begin
		MekRepairFuel := CheckAndDeplete_RepairFuel( NPC_Mek , Skill , 0 , '' , msg );
	end;
	if ( NIL <> NPC_Pilot ) then begin
		PilotRepairFuel := CheckAndDeplete_RepairFuel( NPC_Pilot , Skill , 0 , '' , msg );
	end;
	TotalRepairFuel := MekRepairFuel + PilotRepairFuel;

	if ( RepairLimit < 0 ) or ( TotalRepairFuel < RepairLimit ) then begin
		RepairLimit := TotalRepairFuel;
	end;
{$ENDIF}

{$IFDEF PATCH_GH}
	N := UseRepairSkill( GB , NPC , Target , Skill , RepairLimit );
{$ELSE PATCH_GH}
	N := UseRepairSkill( GB , NPC , Target , Skill );
{$ENDIF PATCH_GH}
{$IFDEF PATCH_I18N}
	msg := ReplaceHash( I18N_MsgString('NPCREPAIR','UseSkill'), GearName(NPC), GearName(Target) );
{$ELSE PATCH_I18N}
	msg := MsgString( 'NPCREPAIR_UseSkill' );
	msg := ReplaceHash( msg , GearName( NPC ) );
	msg := ReplaceHash( msg , GearName( Target ) );
{$ENDIF PATCH_I18N}

	{ Inform the user of the success. }
	if N > 0 then begin
{$IFDEF PATCH_I18N}
		msg := msg + ' ' + ReplaceHash( MsgString( 'NPCREPAIR_Success' ) , BStr( N ) );
{$ELSE PATCH_I18N}
		msg := msg + ' ' + MsgString( 'NPCREPAIR_Success' );
		msg := ReplaceHash( msg , BStr( N ) );
{$ENDIF PATCH_I18N}
	end else begin
		msg := msg + ' ' + MsgString( 'NPCREPAIR_Failure' );
	end;

	DialogMsg( msg );

{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
	{ Deplete the fuel. }
	if ( 0 < MekRepairFuel ) then begin
		msg := '';
		CheckAndDeplete_RepairFuel( NPC_Mek , Skill , N , MsgString( 'PCREPAIR_FuelUsedUp' ) , msg );
		if ( '' <> msg ) then begin
			DialogMsg( msg );
		end;
	end;
	if ( ( MekRepairFuel < N ) and ( 0 < PilotRepairFuel ) ) then begin
		msg := '';
		CheckAndDeplete_RepairFuel( NPC_Pilot , Skill , ( N - MekRepairFuel ) , MsgString( 'PCREPAIR_FuelUsedUp' ) , msg );
		if ( '' <> msg ) then begin
			DialogMsg( msg );
		end;
	end;

	AIRepair := N;
{$ELSE}
	RepairFuel^.V := RepairFuel^.V - N;
	if RepairFuel^.V < 1 then begin
		if IsSubCom( RepairFuel ) then begin
			RemoveGear( RepairFuel^.Parent^.SubCom , RepairFuel );
		end else if IsInvCom( RepairFuel ) then begin
			RemoveGear( RepairFuel^.Parent^.InvCom , RepairFuel );
		end;
	end;
{$ENDIF}
end;

Function SelectRepairTarget( GB: GameBoardPtr; Mek: GearPtr; Skill: Integer ): GearPtr;
	{ Locate a target that needs repairs. }
var
	T,BTar: GearPtr;
{$IFDEF PATCH_GH}
	Best,Dmg: LongInt;
{$ELSE PATCH_GH}
	Best,Dmg: Integer;
{$ENDIF PATCH_GH}
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit(NIL);
{$ENDIF PATCH_GH}

	T := GB^.Meks;
	BTar := Nil;
	Best := 0;

	while T <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < T^.G) then begin
{$ENDIF PATCH_GH}
		if AreAllies( GB , Mek , T ) then begin
			Dmg := TotalRepairableDamage( T , Skill );
			if Dmg > Best then begin
				BTar := T;
				Best := Dmg;
			end;
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		T := T^.Next;
	end;

	SelectRepairTarget := BTar;
end;

Function SelectSocialTarget( GB: GameBoardPtr; NPC: GearPtr; MustBeSexy: Boolean ): GearPtr;
	{ Select a character for NPC to interact with. }
var
	M,Target: GearPtr;
	N,Team: Integer;
	Function IsGoodSocTarget: Boolean;
		{ Return TRUE if M is a good target, or FALSE otherwise. }
	begin
{$IFDEF PATCH_GH}
		if (NIL = M) or (M^.G <= GG_DisposeGear) then Exit(False);
{$ENDIF PATCH_GH}
		if ( NAttValue( NPC^.NA , NAG_Personal , NAS_CID ) = 0 ) or ( NAttValue( M^.NA , NAG_Location , NAS_Team ) = Team ) or ( NAttValue( M^.NA , NAG_Location , NAS_Team ) = NAV_DefPlayerTeam ) then begin
			IsGoodSocTarget := False;
		end else if MustBeSexy then begin
			IsGoodSocTarget := GearActive( M ) and ( not AreEnemies( GB , NPC , M ) ) and IsSexy( NPC , M );
		end else begin
			IsGoodSocTarget := GearActive( M ) and ( not AreEnemies( GB , NPC , M ) );
		end;
	end;
begin
{$IFDEF PATCH_GH}
	if (NIL = NPC) or (NPC^.G <= GG_DisposeGear) then Exit(NIL);
{$ENDIF PATCH_GH}

	{ Make two passes. On the first pass, just count the number of candidates. }
	{ On the second pass actually select one. }
	N := 0;
	Target := Nil;
	M := GB^.Meks;
	Team := NAttValue( NPC^.NA , NAG_Location , NAS_Team );
	while M <> Nil do begin
{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		if IsGoodSocTarget then Inc( N );
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}
		M := M^.Next;
	end;

	{ If any potential targets were found, pick one of them at random. }
	if N > 0 then begin
		M := GB^.Meks;
		N := Random( N );
		while M <> Nil do begin
{$IFDEF PATCH_GH}
			if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
			if IsGoodSocTarget then begin
				Dec( N );
				if N = -1 then Target := M;
			end;
{$IFDEF PATCH_GH}
			end;
{$ENDIF PATCH_GH}
			M := M^.Next;
		end;
	end;

	SelectSocialTarget := Target;
end;

Procedure NPC_Flirtation( GB: GameBoardPtr; NPC , Target: GearPtr );
	{ NPC will flirt with TARGET. If successful, this will cause the PC to gain }
	{ a reaction bonus from TARGET. }
var
	skRoll: Integer;
	CID: LongInt;
	msg: String;
	M,PC: GearPtr;
begin
{$IFDEF PATCH_GH}
	if (NIL = NPC) or (NPC^.G <= GG_DisposeGear) then Exit;
	if (NIL = Target) or (Target^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	SkRoll := RollStep( SkillValue( NPC , 27 ) );
	if SkRoll > 15 then begin
		{ Report the success. }
{$IFDEF PATCH_I18N}
		msg := ReplaceHash( I18N_MsgString('NPCFLIRT','Good'), PilotName(NPC), PilotName(TARGET) );
{$ELSE PATCH_I18N}
		msg := MsgString( 'NPCFLIRT_Good' );
		msg := ReplaceHash( msg , PilotName( NPC ) );
		msg := ReplaceHash( msg , PilotName( TARGET ) );
{$ENDIF PATCH_I18N}
		DialogMsg( msg );

		{ Success! Improve the reaction score. }
		CID := NAttValue( Target^.NA , NAG_Personal , NAS_CID );
		M := GB^.Meks;
		while M <> Nil do begin
{$IFDEF PATCH_GH}
			if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
			if ( NAttValue( M^.NA , NAG_Location , NAS_Team ) = NAV_DefPlayerTeam ) and GearActive( M ) then begin
				PC := LocatePilot( M );
				if PC <> Nil then AddNAtt( PC^.NA , NAG_ReactionScore , CID , 1 + Random( 3 ) );
			end;
{$IFDEF PATCH_GH}
			end;
{$ENDIF PATCH_GH}
			M := M^.Next;
		end;
	end else if SkRoll > 5 then begin
		{ Okay... neither good nor bad. }
{$IFDEF PATCH_I18N}
		msg := ReplaceHash( I18N_MsgString('NPCFLIRT','Okay'), PilotName(NPC), PilotName(TARGET) );
{$ELSE PATCH_I18N}
		msg := MsgString( 'NPCFLIRT_Okay' );
		msg := ReplaceHash( msg , PilotName( NPC ) );
		msg := ReplaceHash( msg , PilotName( TARGET ) );
{$ENDIF PATCH_I18N}
		DialogMsg( msg );
	end else begin
		{ Bad. This is just bad. }
		AddMoraleDmg( NPC , 15 );
{$IFDEF PATCH_I18N}
		msg := ReplaceHash( I18N_MsgString('NPCFLIRT','Bad'), PilotName(NPC), PilotName(TARGET) );
{$ELSE PATCH_I18N}
		msg := MsgString( 'NPCFLIRT_Bad' );
		msg := ReplaceHash( msg , PilotName( TARGET ) );
		msg := ReplaceHash( msg , PilotName( NPC ) );
{$ENDIF PATCH_I18N}
		DialogMsg( msg );
	end;
end;

Procedure NPC_Chatting( GB: GameBoardPtr; NPC , Target: GearPtr );
	{ NPC will chat with TARGET. This may reveal a rumor. }
var
	skRoll: Integer;
	msg: String;
	Rumors: SAttPtr;
begin
{$IFDEF PATCH_GH}
	if (NIL = NPC) or (NPC^.G <= GG_DisposeGear) then Exit;
	if (NIL = Target) or (Target^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	SkRoll := RollStep( SkillValue( NPC , 19 ) );
	if SkRoll > 10 then begin
		{ A rumor has been gained. }
		rumors := CreateRumorList( GB , Nil , Target );

		if rumors <> Nil then begin
{$IFDEF PATCH_I18N}
			msg := ReplaceHash( MsgString( 'NPCCHAT_Good' ) , PilotName( NPC ) );
{$ELSE PATCH_I18N}
			msg := MsgString( 'NPCCHAT_Good' );
			msg := ReplaceHash( msg , PilotName( NPC ) );
{$ENDIF PATCH_I18N}
			msg := msg + ' ' + SelectRandomSAtt( Rumors )^.info;
			DisposeSAtt( Rumors );
		end else begin
{$IFDEF PATCH_I18N}
			msg := ReplaceHash( I18N_MsgString('NPCCHAT','Okay'), PilotName(NPC), PilotName(TARGET) );
{$ELSE PATCH_I18N}
			msg := MsgString( 'NPCCHAT_Okay' );
			msg := ReplaceHash( msg , PilotName( NPC ) );
			msg := ReplaceHash( msg , PilotName( TARGET ) );
{$ENDIF PATCH_I18N}
		end;

		DialogMsg( msg );
	end else begin
		{ Okay... neither good nor bad. }
{$IFDEF PATCH_I18N}
		msg := ReplaceHash( I18N_MsgString('NPCCHAT','Okay'), PilotName(NPC), PilotName(TARGET) );
{$ELSE PATCH_I18N}
		msg := MsgString( 'NPCCHAT_Okay' );
		msg := ReplaceHash( msg , PilotName( NPC ) );
		msg := ReplaceHash( msg , PilotName( TARGET ) );
{$ENDIF PATCH_I18N}
		DialogMsg( msg );
	end;
end;

Procedure LancemateUsefulAction( Mek: GearPtr; GB: GameBoardPtr );
	{ Do something useful! There are no enemies to be found... }
const
	Num_AI_Repair = 5;
	AI_Repair_List: Array [1..Num_AI_Repair] of Byte = (
		{ Order of preference: Medical, First Aid, General Repair, Mecha Repair, Biotech }
		16, 20 , 23, 15, 22
	);
var
	HM,Target,T: Integer;
{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
	dummy_msg: String;
	NPC,TGear: GearPtr;
	NPC_Mek: GearPtr;
	MekRepairFuel, PilotRepairFuel, TotalRepairFuel: LongInt;
{$ELSE}
	NPC,TGear,Tool: GearPtr;
{$ENDIF}
	CORD: Integer;	{ Continuous Orders }
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ See if there are any pending actions. }
	CORD := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders );

	{ If CORD < 0 , use a repair skill. }
	if CORD < 0 then begin
		Target := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ATarget );
		TGear := LocateMekByUID( GB , Target );
{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
		NPC_Mek := FindRoot( Mek );
		NPC := LocatePilot( Mek );
		MekRepairFuel := 0;
		PilotRepairFuel := 0;
		if ( ( NIL <> NPC_Mek ) and ( NPC_Mek <> NPC ) ) then begin
			MekRepairFuel := CheckAndDeplete_RepairFuel( NPC_Mek , Abs( Cord ) , 0 , '' , dummy_msg );
		end;
		if ( NIL <> NPC ) then begin
			PilotRepairFuel := CheckAndDeplete_RepairFuel( NPC , Abs( Cord ) , 0 , '' , dummy_msg );
		end;
		TotalRepairFuel := MekRepairFuel + PilotRepairFuel;
{$ELSE}
		Tool := SeekGear( Mek , GG_RepairFuel , Abs( Cord ) );
{$ENDIF}

		{ If this gear no longer needs repairs, quit. }
{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
		if ( TGear = Nil ) or ( TotalRepairableDamage( TGear , Abs( CORD ) ) < 1 ) or ( TotalRepairFuel <= 0 ) or ( CharCurrentMental( NPC ) < 1 ) then begin
{$ELSE}
		if ( TGear = Nil ) or ( TotalRepairableDamage( TGear , Abs( CORD ) ) < 1 ) or ( Tool = Nil ) or ( CharCurrentMental( Mek ) < 1 ) then begin
{$ENDIF}
			{ Clear the continuous action. }
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );

		end else if OnTheMap( TGear ) then begin
			{ If on the map, NPC must move towards TARGET. }
			if Range( GB , Mek , TGear ) > 3 then begin
				HM := GetHotMap( GB , Target , HotMoveMode( Mek ) , ORD_SeekSingleModel );
				MoveTowardsGoal( GB , Mek , HM );
			end else begin
{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
				AIRepair( GB , Mek , TGear , -1 , Abs( CORD ) );
{$ELSE}
				AIRepair( GB , Mek , TGear , Tool , Abs( CORD ) );
{$ENDIF}
			end;

		end else begin
			{ If off the map, NPC can just autorepair it. }
{$IF DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
			AIRepair( GB , Mek , TGear , -1 , Abs( CORD ) );
{$ELSE}
			AIRepair( GB , Mek , TGear , Tool , Abs( CORD ) );
{$ENDIF}
		end;

	end else if CORD = CORD_Flirt then begin
		Target := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ATarget );
		TGear := LocateMekByUID( GB , Target );

		{ If this gear no longer qualifies, quit. }
		if ( TGear = Nil ) or AreEnemies( GB , Mek , TGear ) or ( Destroyed( TGear ) ) then begin
			{ Clear the continuous action. }
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );

		end else if OnTheMap( TGear ) then begin
			{ If on the map, NPC must move towards TARGET. }
			if Range( GB , Mek , TGear ) > 3 then begin
				{ If the move cannot proceed, cancel the action. }
				HM := GetHotMap( GB , Target , HotMoveMode( Mek ) , ORD_SeekSingleModel );
				if not XMoveTowardsGoal( GB , Mek , HM , 0 , 0 ) then SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );
			end else begin
				{ Close enough to flirt. Yay! }
				NPC_Flirtation( GB, Mek , TGEar );
				WaitAMinute( GB , Mek , ReactionTime( Mek ) * 5 );
				SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );
			end;

		end else begin
			{ If off the map, cancel the action. }
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );
		end;

	end else if CORD = CORD_Chat then begin
		Target := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ATarget );
		TGear := LocateMekByUID( GB , Target );

		{ If this gear no longer qualifies, quit. }
		if ( TGear = Nil ) or AreEnemies( GB , Mek , TGear ) or ( Destroyed( TGear ) ) then begin
			{ Clear the continuous action. }
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );

		end else if OnTheMap( TGear ) then begin
			{ If on the map, NPC must move towards TARGET. }
			if Range( GB , Mek , TGear ) > 3 then begin
				{ If the move cannot proceed, cancel the action. }
				HM := GetHotMap( GB , Target , HotMoveMode( Mek ) , ORD_SeekSingleModel );
				if not XMoveTowardsGoal( GB , Mek , HM , 0 , 0 ) then SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );
			end else begin
				{ Close enough to flirt. Yay! }
				NPC_Chatting( GB, Mek , TGEar );
				WaitAMinute( GB , Mek , ReactionTime( Mek ) * 5 );
				SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );
			end;

		end else begin
			{ If off the map, cancel the action. }
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , 0 );
		end;

	end else begin
		{ Attempt to find a useful thing to do. }
		{ Locate the NPC proper. }
{$IFDEF PATCH_GH}
		NPC_Mek := FindRoot( Mek );
{$ENDIF PATCH_GH}
		NPC := LocatePilot( Mek );

		{ Check to see whether or not this NPC has any repair skills }
		{ to use. }
		t := 1;
		Target := 0;
		CORD := 0;
{$IFDEF PATCH_CHEAT}
		if ( 0 < CharCurrentMental( Mek ) ) or ( Cheat_NPConMechaWillRepair and ( 0 < CharCurrentMental( NPC ) ) ) then begin
{$ELSE PATCH_CHEAT}
		if CharCurrentMental( Mek ) > 0 then begin
{$ENDIF PATCH_CHEAT}
			while ( t <= Num_AI_Repair ) and ( Target = 0 ) do begin
{$IFDEF PATCH_GH}
				MekRepairFuel := 0;
				PilotRepairFuel := 0;
				if ( ( NIL <> NPC_Mek ) and ( NPC_Mek <> NPC ) ) then begin
					MekRepairFuel := CheckAndDeplete_RepairFuel( NPC_Mek , AI_Repair_List[t] , 0 , '' , dummy_msg );
				end;
				if ( NIL <> NPC ) then begin
					PilotRepairFuel := CheckAndDeplete_RepairFuel( NPC , AI_Repair_List[t] , 0 , '' , dummy_msg );
				end;
				TotalRepairFuel := MekRepairFuel + PilotRepairFuel;

				if ( 0 < NAttValue( NPC^.NA , NAG_Skill , AI_Repair_List[ t ] ) ) and ( 0 < TotalRepairFuel ) then begin
{$ELSE PATCH_GH}
				if ( NAttValue( NPC^.NA , NAG_Skill , AI_Repair_List[ t ] ) > 0 ) and ( SeekGear( Mek , GG_RepairFuel , AI_Repair_List[ t ] ) <> Nil ) then begin
{$ENDIF PATCH_GH}
					TGear := SelectRepairTarget( GB , Mek , AI_Repair_List[ t ] );
					if TGear <> Nil then begin
						Target := NAttValue( TGear^.NA , NAG_EpisodeData , NAS_UID );
						CORD := -AI_Repair_List[ t ];
					end;
				end;
				Inc( T );
			end;
		end;

		{ If no repair skills, try flirtation or conversation. }
		if ( CORD = 0 ) and ( NAttValue( NPC^.NA , NAG_Skill , 27 ) > Random( 10 ) ) and ( Random( 3 ) = 1 ) and IsSafeArea( GB ) then begin
			TGEar := SelectSocialTarget( GB , NPC , True );
			if TGear <> Nil then begin
				Target := NAttValue( TGear^.NA , NAG_EpisodeData , NAS_UID );
				CORD := CORD_Flirt;
			end;
		end;
		if ( CORD = 0 ) and ( NAttValue( NPC^.NA , NAG_Skill , 19 ) > Random( 10 ) ) and ( Random( 3 ) = 1 ) and IsSafeArea( GB ) then begin
			TGEar := SelectSocialTarget( GB , NPC , False );
			if TGear <> Nil then begin
				Target := NAttValue( TGear^.NA , NAG_EpisodeData , NAS_UID );
				CORD := CORD_Chat;
			end;
		end;

		if CORD <> 0 then begin
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ContinuousOrders , CORD );
			SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_ATarget , Target );
		end else begin
			HM := GetHotMap( GB , NAV_DefPlayerTeam , HotMoveMode( Mek ) , ORD_SeekTeam );
			if not XMoveTowardsGoal( GB , Mek , HM , 5 , 10 ) then Wander( Mek , GB );
		end;
	end;
end;

Procedure Seek_And_Destroy( Mek: GearPtr; GB: GameBoardPtr );
	{ Seek an enemy, destroy it if possible. }

{$IFDEF PATCH_CHEAT}
	Function CheckToFire( Mek: GearPtr; GB: GameBoardPtr ): Boolean;
	var
		Weapon: GearPtr;
		TL,Target: GearPtr;

		procedure SeekFarWeapon( Part: GearPtr );
		begin
			while ( NIL <> Part ) do begin
				if ( ( GG_Module = Part^.G ) or ( GG_Weapon = Part^.G ) ) or ( Cheat_NPC_UseGrenade and ( GG_Ammo = Part^.G ) ) then begin
{$IFDEF PATCH_GH}
					if NeedAmmo( Part ) then begin
						ReloadAmmo( GB , Mek , Part );
					end;
{$ENDIF PATCH_GH}
					if ReadyToFire( GB , Mek , Part , False , True, False ) then begin
						if ( NIL = Weapon ) then begin
							Weapon := Part
						end else if ( WeaponRange( GB , Weapon ) < WeaponRange( GB , Part ) ) then begin
							Weapon := Part;
						end;
					end;
				end;
				SeekFarWeapon( Part^.SubCom );
				SeekFarWeapon( Part^.InvCom );
				Part := Part^.Next;
			end;
		end;
	begin
		Weapon := NIL;
		SeekFarWeapon( Mek^.SubCom );
		SeekFarWeapon( Mek^.InvCom );
		if ( NIL = Weapon ) then begin
			exit( False );
		end;

		Target := NIL;
		TL := GB^.Meks;

		while ( NIL <> TL ) do begin
			if (GG_DisposeGear < TL^.G) then begin
				if ( AreEnemies( GB , Mek , TL ) and OnTheMap( TL ) ) then begin
					if ( RangeCheck( GB , Mek , Weapon , TL ) and GearActive( TL ) ) then begin
						if MekCanSeeTarget( GB , Mek , TL ) then begin
							Target := TL;
							break;
						end;
					end;
				end;
			end;
			TL := TL^.Next;
		end;

		if ( NIL = Target ) then begin
			exit( False );
		end;

		CheckToFire := True;
	end;
{$ENDIF PATCH_CHEAT}

var
	HM: Integer;
	N1,D1,N2,D2: LongInt;	{ Numerator 1 , Numerator 2 , Denominator 1 , Denominator 2 }
				{ Used for the OptMax, OptMin calculations. }
	Procedure CheckOptimumRange( Part: GearPtr );
		{ Check the weapons along this track to find out the optimum fighting }
		{ ranges for MEK. }
		{ This is done with a strange fraction calculation- find the average range }
		{ of all weapons, weighted by the DC they do. }
	var
		Dmg,Rng: Integer;
	begin
		while Part <> Nil do begin
{$IFDEF PATCH_GH}
			if (GG_DisposeGear < Part^.G) then begin
{$ENDIF PATCH_GH}
			{ If PART is a weapon, it will affect our calculations. }
{$IFDEF PATCH_GH}
			if NeedAmmo( Part ) then begin
				ReloadAmmo( GB , Mek , Part );
			end;
{$ENDIF PATCH_GH}
			if ReadyToFire( GB , Mek , Part ) then begin
				if Part^.G = GG_Module then begin
					Dmg := WeaponDC( Part , 0 );
					D1 := D1 + Dmg div 2;
				end else if Part^.G = GG_Weapon then begin
					{ Find the DMG and RNG of this weapon. }
					Dmg := WeaponDC( Part , 0 );
					Rng := WeaponRange( GB , Part );
					if Rng < 2 then begin
						D1 := D1 + Dmg div 2;
						D2 := D2 + Dmg * 6;
					end else if Rng < 18 then begin
						N1 := N1 + ( Rng * 2 * Dmg );
						N2 := N2 + ( Rng * Dmg div 3 );
						D1 := D1 + Dmg;
						D2 := D2 + Dmg;
					end else begin
						N1 := N1 + ( Rng * 2 * Dmg );
						D1 := D1 + Dmg;
						N2 := N2 + ( Rng * Dmg );
						D2 := D2 + Dmg;
					end;
{$IFDEF PATCH_CHEAT}
				end else if GG_Ammo = Part^.G then begin
					Dmg := WeaponDC( Part , 0 );
					D1 := D1 + Dmg div 2;
{$ENDIF PATCH_CHEAT}
				end;
			end;
			CheckOptimumRange( Part^.SubCom );
			CheckOptimumRange( Part^.InvCom );
{$IFDEF PATCH_GH}
			end;
{$ENDIF PATCH_GH}
			Part := Part^.Next;
		end;
	end;
	Procedure GetOptimumRange;
		{ Determine the optimum range for this mecha. }
	var
		NPC: GearPtr;
	begin
		NPC := LocatePilot( Mek );

		if ( NPC <> Nil ) and ( NAttValue( NPC^.NA , NAG_Personal , NAS_OptMax ) <> 0 ) then begin
			N1 := NAttValue( NPC^.NA , NAG_Personal , NAS_OptMax );
			N2 := NAttValue( NPC^.NA , NAG_Personal , NAS_OptMin );
		end else begin
			CheckOptimumRange( Mek^.SubCom );
		end;
	end;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ First, move towards the enemy, if necessary. }
	if Random( 3 ) = 1 then begin
		SelectMoveMode( Mek , GB );
	end;
	HM := GetHotMap( GB , NAttValue( Mek^.NA , NAG_LOcation , NAS_Team ) , HotMoveMode( Mek ) , ORD_SeekEnemy );

	{ Determine optimum ranges. }
	N1 := 0;
	N2 := 0;
	D1 := 1;
	D2 := 1;
	GetOptimumRange;
	if not XMoveTowardsGoal( GB , Mek , HM , N1 div D1 , N2 div D2 ) then begin
		if NAttValue( Mek^.NA , NAG_Location , NAS_Team ) = NAV_LancemateTeam then begin
			LancemateUsefulAction( Mek , GB );
		end else begin
			Wander( Mek , GB );
		end;
	end else begin
		if ( GB^.ComTime > NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ChatterRecharge ) ) and ( Random( 50 ) = 1 ) then begin
			NPC_CombatTaunt( GB , Mek , 'CHAT_ATTACK' );
		end;
	end;

	{ Secondly, attack anyone within reach. }
	AttackTargetOfOppurtunity( GB , Mek );

{$IFDEF PATCH_CHEAT}
	if Random( 10 ) = 0 then begin
		if CheckToFire( Mek, GB ) then begin
			SelectHeavyFormMode( Mek , GB );
		end else begin
{$IFDEF PATCH_CHEAT}
			SelectPurgePartsMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
			SelectFastestFormMode( Mek , GB );
		end;
	end;
{$ENDIF PATCH_CHEAT}
end;


Procedure SetNewOrders( GB: GameBoardPtr; Mek: GearPtr );
	{ MEK has apparently either completed what it set out to do, }
	{ or the standing orders have become impossible. }
var
	TN: Integer;
	TG: GearPtr;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	TN := NAttValue( Mek^.NA , NAG_Location , NAS_Team );
	TG := LocateTeam( GB , TN );

	if TG <> Nil then begin
		{ Set the default orders listed in the TEAM gear. }
		SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_Orders , TG^.Stat[ STAT_TeamOrders ] );

	end else begin
		{ If no team gear has been defined, use default orders }
		{ SEARCH AND DESTROY. }
		SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_Orders , NAV_SeekAndDestroy );
	end;
end;

Procedure GOTO_SPOT( Mek: GearPtr; GB: GameBoardPtr );
	{ The outline for this behavior is as follows: }
	{ - If not moving, set move towards target. }
	{ - Fire at targets of oppurtunity. }
	{ - Once target is reached, seek new orders. }
var
	HM,X,Y,GX,GY: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Locate all the values we're gonna need. }
	X := NAttValue( Mek^.NA , NAG_Location , NAS_X );
	Y := NAttValue( Mek^.NA , NAG_Location , NAS_Y );
	GX := NAttValue( Mek^.NA , NAG_Location , NAS_GX );
	GY := NAttValue( Mek^.NA , NAG_Location , NAS_GY );

	if Random( 3 ) = 1 then begin
{$IFDEF PATCH_CHEAT}
		SelectPurgePartsMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_CHEAT}
		SelectFastestFormMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
		SelectMoveMode( Mek , GB );
	end;

	{ If we have reached the spot where we want to go, }
	{ its time to seek new orders. }
	if ( GX = X ) and ( GY = Y ) then begin
		SetNewOrders( GB , Mek );

	end else begin
		HM := GetHotMap( GB , GX + ( GY * ( XMax + 1 ) ) , HotMoveMode( Mek ) , ORD_SeekSpot );
		MoveTowardsGoal( GB , Mek , HM );

		{ Secondly, attack an enemy within reach. }
		AttackTargetOfOppurtunity( GB , Mek );
	end;
end;

Procedure GOTO_EDGE( Mek: GearPtr; GB: GameBoardPtr );
	{ The outline for this behavior is as follows: }
	{ - If not moving, set move towards target. }
	{ - Fire at targets of oppurtunity. }
var
	HM,Edge: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Locate all the values we're gonna need. }
	Edge := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ATarget );

	if Random( 2 ) = 1 then begin
{$IFDEF PATCH_CHEAT}
		SelectPurgePartsMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_CHEAT}
		SelectFastestFormMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
		SelectMoveMode( Mek , GB );
	end;

	HM := GetHotMap( GB , Edge , HotMoveMode( Mek ) , ORD_SeekEdge );
	MoveTowardsGoal( GB , Mek , HM );

	{ Secondly, attack an enemy within reach. }
	AttackTargetOfOppurtunity( GB , Mek );
end;

Procedure PASSIVE( Mek: GearPtr; GB: GameBoardPtr );
	{ This particular gear is going to act passively. It will not }
	{ move around much, just stand there mostly. If it spots a }
	{ hostile model it will run about randomly. }
	{ This AI type is to be used for townsfolk and other unfortunates. }
var
	HM: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ First, move towards the enemy, if necessary. }
	HM := GetHotMap( GB , NAttValue( Mek^.NA , NAG_LOcation , NAS_Team ) , HotMoveMode( Mek ) , ORD_SeekEnemy );
	FleeFromGoal( GB , Mek , HM );
end;

Procedure RUNAWAY( Mek: GearPtr; GB: GameBoardPtr );
	{ This AI type is for models wishing to exit the board. }
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ This mode will likely run the NPC off the map. }
	if MoveBlocked( Mek , GB ) then begin
{$IFDEF PATCH_CHEAT}
		SelectPurgePartsMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
{$IFDEF PATCH_CHEAT}
		SelectFastestFormMode( Mek , GB );
{$ENDIF PATCH_CHEAT}
		SelectMoveMode( Mek , GB );
		if Random( 2 ) = 1 then begin
			PrepAction( GB , Mek , NAV_TurnRight );
		end else begin
			PrepAction( GB , Mek , NAV_TurnLeft );
		end;
	end else begin
		PrepAction( GB , Mek , NAV_FullSpeed );
	end;
end;

Procedure FOLLOW( Mek: GearPtr; GB: GameBoardPtr );
	{ Attempt to follow the target. }
	{ Note that this is a "MAGIC" AI type, since the NPC will }
	{ follow its target even if the target cannot be seen. }
var
	UID,HM: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ First, attempt to move closer to target. }
	UID := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ATarget );
	HM := GetHotMap( GB , UID , HotMoveMode( Mek ) , ORD_SeekSingleModel );
	MoveTowardsGoal( GB , Mek , HM );

	{ Secondly, attack an enemy within reach. }
	AttackTargetOfOppurtunity( GB , Mek );
end;

Procedure AI_Eject( Mek: GearPtr; GB: GameBoardPtr );
	{ This NPC is ejecting from his mecha! }
var
	Pilot: GearPtr;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Better set the following triggers. }
	SetTrigger( GB , TRIGGER_NumberOfUnits + BStr( NAttValue( Mek^.NA , NAG_Location , NAS_Team ) ) );
	SetTrigger( GB , TRIGGER_UnitEliminated2 + BStr( NAttValue( Mek^.NA , NAG_EpisodeData , NAS_UID ) ) );

	repeat
		Pilot := ExtractPilot( Mek );

		if Pilot <> Nil then begin
			NPC_CombatTaunt( GB , Pilot , 'CHAT_EJECT' );
			DialogMsg( ReplaceHash( MsgString( 'EJECT_AI' ) , GearName( Pilot ) ) );
			DeployMek( GB , Pilot , False );
		end;
	until Pilot = Nil;
end;

Function ShouldEject( Mek: GearPtr; GB: GameBoardPtr ): Boolean;
	{ Return TRUE if this mecha should eject, or FALSE otherwise. }
var
	Dmg,PrevDmg,Intimidation: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit(False);
{$ENDIF PATCH_GH}

	Dmg := PercentDamaged( Mek );
	PrevDmg := 100 - NAttValue( Mek^.NA , NAG_EpisodeData , NAS_PrevDamage );
	SetNAtt( Mek^.NA , NAG_EpisodeData , NAS_PrevDamage , 100 - DMG );
	if AreEnemies( GB , NAttValue( Mek^.NA , NAG_Location , NAS_Team ) , NAV_DefPlayerTeam ) and ( Dmg < 75 ) and ( DMG < PrevDmg ) then begin
		Intimidation := TeamSkill( GB , NAV_DefPlayerTeam , 28 );
		if BaseMoveRate( Mek ) = 0 then Dmg := Dmg - RollStep( Intimidation ) - RollStep( Intimidation );
		ShouldEject := Dmg < ( Random( 45 ) + RollStep( Intimidation ) - RollStep( SkillValue( Mek , 28 ) ) );
	end else if AreAllies( GB , NAttValue( Mek^.NA , NAG_Location , NAS_Team ) , NAV_DefPlayerTeam ) and ( Dmg < 75 ) and ( DMG < PrevDmg ) then begin
		{ Determine the player team's Leadership score. }
		Intimidation := TeamSkill( GB , NAV_DefPlayerTeam , 39 );
		if BaseMoveRate( Mek ) = 0 then Dmg := Dmg - 25;
		ShouldEject := Dmg < ( Random( 60 ) - RollStep( Intimidation ) );
	end else ShouldEject := False;
end;

Procedure GetAIInput( Mek: GearPtr; GB: GameBoardPtr );
	{ MEK belongs to the computer team. Decide upon }
	{ a course of action for it here. }
var
	O: LongInt;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Before processing orders, check jump time, since the AI is }
	{ stupid and will crash as often as possible. }
	O := NAttValue( Mek^.NA , NAG_Action , NAS_TimeLimit );
	if ( O > 0 ) and (( O + CalcMoveTime( Mek , GB ) ) >= GB^.ComTime ) then begin
		{ This is as far as this jump can go. Better land now. }
		PrepAction( GB , Mek , NAV_Stop );

		{ Switch to another movemode, since jumping will }
		{ require some time to recharge. }
		GearUp( Mek );
	end;

	{ If MEK is actually a mecha, and it's significantly damaged, }
	{ maybe the pilot would rather eject than continue fighting. }
	if ( Mek^.G = GG_Mecha ) and ShouldEject( Mek , GB ) then begin
		AI_EJECT( Mek , GB );

	end else begin

		{ Otherwise, check the orders of the NPC unit, and branch accordingly. }
		O := NAttValue( Mek^.NA , NAG_EpisodeData , NAS_Orders );

		case O of
			NAV_GotoSpot:		GOTO_SPOT( Mek , GB );
			NAV_SeekEdge:		GOTO_EDGE( Mek , GB );
			NAV_Passive:		PASSIVE( Mek , GB );
			NAV_RunAway:		RUNAWAY( Mek , GB );
			NAV_Follow:		FOLLOW( Mek , GB );
			else Seek_And_Destroy( Mek , GB );
		end;

		if ( GB^.ComTime > NAttValue( Mek^.NA , NAG_EpisodeData , NAS_ChatterRecharge ) ) and ( Random( 80 ) = 1 ) and IsSafeArea( GB ) then begin
			NPC_CombatTaunt( GB , Mek , 'CHAT_SAFE' );
		end;
	end;
end;

Procedure ConfusedInput( Mek: GearPtr; GB: GameBoardPtr );
	{ The mek is not in full control of its faculties. Move randomly. }
var
	P: Point;
	CD: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Determine current facing and position. }
	P := GearCurrentLocation( Mek );
	CD := NAttValue( Mek^.NA , NAG_Location , NAS_D );

	{ If the current direction of travel doesn't lead off the map, }
	{ and isn't blocked, just move foreword. }
	if OnTheMap( P.X + AngDir[ CD , 1 ] , P.Y + AngDir[ CD , 2 ] ) and ( Random( 3 ) <> 1 ) and not MoveBlocked( Mek , GB ) then begin
		if Random( 2 ) = 1 then begin
			PrepAction( GB , Mek , NAV_FullSpeed );
		end else begin
			PrepAction( GB , Mek , NAV_NormSpeed );
		end;
	end else begin
		if Random( 2 ) = 1 then begin
			PrepAction( GB , Mek , NAV_TurnRight );
		end else begin
			PrepAction( GB , Mek , NAV_TurnLeft );
		end;
	end;
end;

Procedure BrownianMotion( GB: GameBoardPtr );
	{ Go through all the clouds and flames on the map, and update }
	{ them. A few rules... }
	{  1. No more than one cloud and one flame per tile. }
	{  2. Clouds will move if there's a free square. }
	{  3. Fires will burn terrain and ay reproduce to free squares. }
	{  4. Both fires and clouds will do EFFECT if that SAtt is defined. }

	Procedure MetaEffect( MT: GearPtr; X,Y: Integer );
		{ This metaterrain has some effect. Better handle that }
		{ here. }
	var
		fx,desc: String;
		N,T: Integer;
		Target: GearPtr;
	begin
		fx := SAttValue( MT^.SA , 'EFFECT' );
		desc := SAttValue( MT^.SA , 'FX_DESC' );
		N := NumGearsXY( GB , X , Y );
		for t := 1 to N do begin
			target := FindGearXY( GB , X , Y , T );
{$IFDEF PATCH_GH}
			if ( Target <> MT ) and NotDestroyed( Target ) then begin
				EffectFrontEnd( GB , MT , Target , fx , desc );
			end;
{$ELSE PATCH_GH}
			if ( Target <> MT ) and NotDestroyed( Target ) then EffectFrontEnd( GB , Target , fx , desc );
{$ENDIF PATCH_GH}
		end;
	end;
var
	CloudMap: Array [1..XMax,1..YMax] of GearPtr;
	FireMap: Array [1..XMax,1..YMax] of GearPtr;
	ElseMap: Array [1..XMax,1..YMax] of Boolean;
	M,M2: GearPtr;
	P: Point;
	X,Y,T: Integer;
begin
	{ Start by filling out the cloudmap and the firemap. }
	for x := 1 to XMax do begin
		for y := 1 to YMax do begin
			CloudMap[ X , Y ] := Nil;
			FireMap[ X , Y ] := Nil;
			ElseMap[ X , Y ] := False;
		end;
	end;

	{ Look for clouds and flames and store them in the map. }
	M := GB^.Meks;
	while M <> Nil do begin
		M2 := M^.Next;

{$IFDEF PATCH_GH}
		if (GG_DisposeGear < M^.G) then begin
{$ENDIF PATCH_GH}
		P := GearCurrentLocation( M );
		if OnTheMap( P.X , P.Y ) then begin
			if ( M^.G = GG_Metaterrain ) and ( M^.S = GS_MetaCloud ) then begin
				{ If two clouds exist on the same tile, }
				{ the one with the greatest duration wins. }
				{ Remember that a negative duration really }
				{ means infinite duration. }
				if CloudMap[ P.X , P.Y ] = Nil then begin
					CloudMap[ P.X , P.Y ] := M;
				end else if M^.Stat[ STAT_CloudDuration ] < 0 then begin
					RemoveGear( GB^.Meks , CloudMap[ P.X , P.Y ] );
					CloudMap[ P.X , P.Y ] := M;
				end else if ( M^.Stat[ STAT_CloudDuration ] > CloudMap[ P.X , P.Y ]^.Stat[ STAT_CloudDuration ] ) and ( CloudMap[ P.X , P.Y ]^.Stat[ STAT_CloudDuration ] > 0 ) then begin
					RemoveGear( GB^.Meks , CloudMap[ P.X , P.Y ] );
					CloudMap[ P.X , P.Y ] := M;
				end else begin
					RemoveGear( GB^.Meks , M );
				end;
			end else if ( M^.G = GG_Metaterrain ) and ( M^.S = GS_MetaFire ) then begin
				{ Fires are simpler than clouds. Just one }
				{ fire per map square, no exceptions. }
				if FireMap[ P.X , P.Y ] = Nil then begin
					FireMap[ P.X , P.Y ] := M;
				end else begin
					RemoveGear( GB^.Meks , M );
				end;

			end else begin
				ElseMap[ P.X , P.Y ] := True;
			end;
		end;
{$IFDEF PATCH_GH}
		end;
{$ENDIF PATCH_GH}

		M := M2;
	end;

	{ Finally, go through each cloud/flame one more time and update }
	{ them. }
	for X := 1 to XMax do begin
		for Y := 1 to YMax do begin
			{ If there's a cloud here, deal with it. }
			if CloudMap[ X , Y ] <> Nil then begin
				{ If the cloud has an effect, apply that }
				{ effect to all models in the same tile. }
				if ( SAttValue( CloudMap[ X , Y ]^.SA , 'EFFECT' ) <> '' ) and ElseMap[ X , Y ] then begin
					MetaEffect( CloudMap[ X , Y ] , X , Y );
				end;

				{ If the cloud has a duration, deal with it. }
				if CloudMap[ X , Y ]^.Stat[ STAT_CloudDuration ] > 0 then Dec( CloudMap[ X , Y ]^.Stat[ STAT_CloudDuration ] );
				if CloudMap[ X , Y ]^.Stat[ STAT_CloudDuration ] = 0 then begin
					RemoveGear( GB^.Meks , CloudMap[ X , Y ] );
					RedrawTile( GB , X , Y );

				{ If the cloud's time isn't finished, }
				{ try moving it around. }
				end else begin
					{ Pick a random spot next to the cloud. }
					{ If it's empty and non-blocking, move }
					{ the cloud there. }
					P.X := X + Random( 2 ) - Random( 2 );
					P.Y := Y + Random( 2 ) - Random( 2 );
					if OnTheMap( P.X , P.Y ) and ( CloudMap[ P.X , P.Y ] = Nil ) and not TileBlocksLOS( GB , P.X , P.Y , 5 ) then begin
						CloudMap[ P.X , P.Y ] := CloudMap[ X , Y ];
						CloudMap[ X , Y ] := Nil;
						SetNAtt( CloudMap[ P.X , P.Y ]^.NA , NAG_Location , NAS_X , P.X );
						SetNAtt( CloudMap[ P.X , P.Y ]^.NA , NAG_Location , NAS_Y , P.Y );
						RedrawTile( GB , X , Y );
						RedrawTile( GB , P.X , P.Y );
					end;
				end;
			end; { if CloudMap... }

			if FireMap[ X , Y ] <> Nil then begin
				{ Handle the effects of fire here. }
				if ( SAttValue( FireMap[ X , Y ]^.SA , 'EFFECT' ) <> '' ) and ElseMap[ X , Y ] then begin
					MetaEffect( FireMap[ X , Y ] , X , Y );
				end;

				{ Fire needs smoke. If there's no smoke, }
				{ probably add some. }
				if ( Random( 3 ) <> 1 ) and ( CloudMap[ X , Y ] = Nil ) then begin
					M := LoadNewSTC( 'SMOKE-1' );
					if M <> Nil then begin
						M^.Scale := GB^.Scale;
						AppendGear( GB^.Meks , M );
						M^.Stat[ STAT_CloudDuration ] := RollStep( 2 );
						SetNAtt( M^.NA , NAG_Location , NAS_X , X );
						SetNAtt( M^.NA , NAG_Location , NAS_Y , Y );
					end;
				end;

				{ Check to see if the fire will spread. }
				for t := 0 to 7 do begin
					P.X := X + AngDir[ t , 1 ];
					P.Y := Y + AngDir[ t , 2 ];

					if OnTheMap( P.X , P.Y ) and TerrMan[ GB^.map[ P.X , P.Y ].terr ].Flammable and ( FireMap[ P.X , P.Y ] = Nil ) and ( Random( 15 ) = 1 ) then begin
						M := LoadNewSTC( 'FIRE-1' );
						if M <> Nil then begin
							AppendGear( GB^.Meks , M );
							M^.Scale := GB^.Scale;
							SetNAtt( M^.NA , NAG_Location , NAS_X , P.X );
							SetNAtt( M^.NA , NAG_Location , NAS_Y , P.Y );
							RedrawTile( GB , P.X , P.Y );
						end;
					end;
				end;

				{ Check to see if the fire will go out. }
				if ( Random( 20 ) = 1 ) or not TerrMan[ GB^.map[ X , Y ].terr ].Flammable then begin
					RemoveGear( GB^.Meks , FireMap[ X , Y ] );
					RedrawTile( GB , X , Y );

				end else if Random( TerrMan[ GB^.map[ X , Y ].terr ].DMG + 2 ) = 1 then begin
					DestroyTerrain( GB , X , Y );
					if ( Random( 10 ) = 1 ) then RemoveGear( GB^.Meks , FireMap[ X , Y ] );
					RedrawTile( GB , X , Y );
				end;
			end; { if FireMap... }

		end; {for Y}
	end; {for X}
end;

Function MOVE_MODEL_TOWARDS_SPOT( Mek: GearPtr; GB: GameBoardPtr; GX,GY: Integer ): Boolean;
	{ A new life for an old procedure. }
var
	HM,X,Y: Integer;
	it: Boolean;
begin
{$IFDEF PATCH_GH}
	if (NIL = Mek) or (Mek^.G <= GG_DisposeGear) then Exit;
{$ENDIF PATCH_GH}

	{ Locate all the values we're gonna need. }
	X := NAttValue( Mek^.NA , NAG_Location , NAS_X );
	Y := NAttValue( Mek^.NA , NAG_Location , NAS_Y );

	if Random( 3 ) = 1 then begin
		SelectMoveMode( Mek , GB );
	end;

	{ If we have reached the spot where we want to go, }
	{ its time to seek new orders. }
	if ( GX = X ) and ( GY = Y ) then begin
		MOVE_MODEL_TOWARDS_SPOT := False;

	{ If we're right next to the target but it's blocked, exit false. }
	end else if ( Abs( GX - X ) <= 1 ) and ( Abs( GY - Y ) <= 1 ) and IsBlocked( Mek , GB , GX , GY ) then begin
		MOVE_MODEL_TOWARDS_SPOT := False;

	end else begin
		HM := GetHotMap( GB , GX + ( GY * ( XMax + 1 ) ) , HotMoveMode( Mek ) , ORD_SeekSpot );

		{ If there's no way from here to the desired location, exit FALSE. }
		if HotMap[ HM , X , Y ] > 1000 then Exit( False );

		{ Otherwise attempt to move. }
		MOVE_MODEL_TOWARDS_SPOT := XMoveTowardsGoal( GB , Mek , HM , 0 , 0 );
	end;
end;



initialization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: aibrain.pp');
{$ENDIF DEBUG}
	NPC_Chatter_Standard := LoadStringList( NPC_Chatter_File );
end;

finalization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: aibrain.pp(finalization)');
{$ENDIF DEBUG}
	DisposeSAtt( NPC_Chatter_Standard );
end;

end.
